/*                                                                                                                                                                                 
 *      NiceShaper - Dynamic Traffic Management                                                                                                                                    
 *                                                                                                                                                                                 
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>                                                                                                              
 *                                                                                                                                                                                 
 *      This file is subject to the terms and conditions of the GNU General Public                                                                                                 
 *      License.  See the file COPYING in the main directory of this archive for                                                                                                   
 *      more details.                                                                                                                                                              
 */   

#include "class.h"
#include "aux.h"

TcClass::TcClass( std::vector < char * > &section_ifaces, std::vector < char * > &mark_ifaces)
{
    SectionIfaces = section_ifaces;
    MarkIfaces = mark_ifaces;
    strncpy( Dev, "", MAX_IFACE+1 );
    strncpy( Name, "", MAX_NAME+1 );
    strcpy( EsfqHash, "classic" );
    Alive = 0;
    NsLow = 0;
    NsCeil = 0;
    HardLimit = 0;
    SoftLimit = 0;
    Rate = 0;
    MeasureRate = 0;
    HtbRate = 0;
    HtbCeil = 0;
    ClassId = 0;
    HtbPrio = 5;
    Hold = 30;
    Prio = 0;
    Strict = 0.70;
    Filter = u32;
    OldHtbCeil = 0;
    OldHtbRate = 0;
    SfqPerturb = 10;
    EsfqPerturb = 10;
    Work = false;
    UseIpt = true;
    UseTcClass = true;
    UseTcFilter = true;
    StatsShowHtbCeil = true;
    StatsShowRate = true;
    SumInSection = true;
    LocalSectionId = section_id;
    SectionShape = 0;
    strcpy( MlBuf, "" );
    strcpy( MsBuf, "" );
}

TcClass::~TcClass()
{
    for ( unsigned int n=0; n<tcfilter.size(); n++ ) delete tcfilter[n];

    tcfilter.clear();
}

int TcClass::store(char buf[])
{
    char option[SHORT_BUF+1] = "", value[SHORT_BUF+1] = "";
    char param[SHORT_BUF+1] = "";
    
    if ( test->validWhiteLine(buf)) return 0;
      
    strncpy( option, awk( buf, 1 ), SHORT_BUF );
    strncpy( param, awk( buf, 2 ), SHORT_BUF );
    strncpy( value, awk( buf, 3 ), SHORT_BUF );
    
    if ( !strcmp( option, "class" )) 
    {
	strncpy( Dev, awk( buf, 3 ), MAX_IFACE );
	strncpy( Name, awk( buf, 4 ), MAX_NAME );
	
	if ( !test->validSysDev(Dev))
	{
	    error->dump ( 16, buf );
	    return -1;
	}
	
	if ( is_in_vector( MarkIfaces, Dev )) Filter = fw; 
	else Filter = u32;
	
	if ( !strlen( Name )) 
	{
	    error->dump ( 24, buf );
	    return -1;
	}
	
	if ( !is_in_vector( SectionIfaces, Dev ))
	{
	    error->dump ( 33, buf );
	    return -1;
	}
    }
    else if ( !strcmp( option, "match" )) { tcfilter.push_back( new TcFilter( buf, Dev, Filter )); }     
    else if ( !strcmp( option, "classid" )) ClassId = atoi( param );
    else if ( !strcmp( option, "low" )) NsLow = unitconvert( param, bit );
    else if ( !strcmp( option, "ceil" )) HtbCeil = NsCeil = unitconvert( param, bit );
    else if ( !strcmp( option, "strict" )) Strict = (double)(atoi(param))/100;
    else if ( !strcmp( option, "prio" )) Prio = atoi( param );
    else if ( !strcmp( option, "rate" )) { NsLow = unitconvert( param, bit ); HtbCeil = NsCeil = unitconvert( param, bit );} 
    else if ( !strcmp( option, "htb" )) {
    	if ( !strcmp( param, "prio" )) {
	    HtbPrio = atoi( value );
	}    
	else if ( !strcmp( param, "scheduler" )) {
	    if ( !strcmp( value, "none" )) Qdisc = none;
	    else if ( !strcmp( value, "sfq" )) Qdisc = sfq;
	    else if ( !strcmp( value, "esfq" )) Qdisc = esfq;
	    else { error->dump ( 17, buf ); return -1; }
	}	
	else {
	    error->dump ( 11, buf );	
	    return -1;
	}        
    }
    else if ( !strcmp( option, "sfq" ))
    {
    	if ( !strcmp( param, "perturb" )) SfqPerturb = atoi( value );  
        else { error->dump( 11, buf ); return -1; }    
    }
    else if ( !strcmp( option, "esfq" )) {
        if ( !strcmp( param, "perturb" )) {
    	    EsfqPerturb = atoi( value );  
    	}
	else if ( !strcmp( param, "hash" )) {
	    if ( !strcmp( value, "classic" )) strcpy( EsfqHash, "classic" );
	    else if ( !strcmp( value, "dst" )) strcpy( EsfqHash, "dst" );
	    else if ( !strcmp( value, "src" )) strcpy( EsfqHash, "src" );
	    else { error->dump( 11, buf ); return -1; }
	}    
	else { error->dump( 11, buf ); return -1; }
    }	
    else if ( !strcmp( option, "hold" )) 
    {
	if ( !strlen( param )) Hold = 0;
	else Hold = atoi(param);
    }
//    else if (!strcmp(option, "show"))
//    {
//	if (!strcmp(value, "stats")) 
//	{
//	    stats_show_htb_ceil = false; stats_show_rate = false;
//	}
//    } 
    else if ( !strcmp(option, "standard-class" ))     
    {
    	StatsShowHtbCeil = true; StatsShowRate = true;
	UseTcClass = true; UseTcFilter = true; SumInSection = true;
    }
    else if ( !strcmp(option, "do-not-shape" )) 
    {
	StatsShowHtbCeil = false; StatsShowRate = true;
	UseTcClass = false; UseTcFilter = true; SumInSection = false;
	NsLow = 0; NsCeil = 0; HtbRate = 0; HtbCeil = 0;
    } 
    else if ( !strcmp(option, "virtual" )) 
    {
	StatsShowHtbCeil = false; StatsShowRate = true;
	UseTcClass = false; UseTcFilter = false; SumInSection = false;
	NsLow = 0; NsCeil = 0; HtbRate = 0; HtbCeil = 0;
    } 
    else if ( !strcmp( option, "wrapper" )) 
    {
	StatsShowHtbCeil = true; StatsShowRate = true;
	UseTcFilter = true; UseTcClass = true; SumInSection = false;
	NsLow = NsCeil;
	LocalSectionId = 1;    
    } 
    else if ( !strcmp( option, "set-mark" )) 
    { } 

    if ( strcmp( option, "match" )) {
	for ( unsigned int i=0; i<tcfilter.size(); i++ ) {
	    if ( tcfilter[i]->store(buf) == -1 ) return -1; 	 
	}    
    }    
    
    if ( NsLow > 1073741824 ) { error->dump(32, buf); return -1; }
    if ( NsCeil > 1073741824 ) { error->dump(32, buf); return -1; }   
    
    return 1;
}

int TcClass::prepareTcClass( unsigned int section_shape, unsigned int section_classes )
{
    unsigned int htb_rate = section_shape / section_classes; 
    
    SectionShape = section_shape;

    if ( NsLow > NsCeil ) NsCeil = NsLow;
    if ( UseTcClass ) 
    {
	if ( !SumInSection ) htb_rate = HtbCeil;
	sprintf( TcClassAdd, "tc class add dev %s parent 1:%d classid 1:%d htb rate %d ceil %d prio %d quantum %d", Dev, LocalSectionId, ClassId, htb_rate, htb_rate, HtbPrio, compute_quantum(htb_rate) );
	sprintf( TcClassDel, "tc class del dev %s classid 1:%d", Dev, ClassId );
	if ( Qdisc == sfq )
	{
	    sprintf( TcQdiscAdd, "tc qdisc add dev %s parent 1:%d sfq perturb %d", Dev, ClassId, SfqPerturb );
	    sprintf( TcQdiscDel, "tc qdisc del dev %s parent 1:%d", Dev, ClassId );
	}
	if ( Qdisc == esfq )
	{
	    sprintf( TcQdiscAdd, "tc qdisc add dev %s parent 1:%d esfq perturb %d hash %s", Dev, ClassId, EsfqPerturb, EsfqHash);
	    sprintf( TcQdiscDel, "tc qdisc del dev %s parent 1:%d", Dev, ClassId );
	}
	else if ( Qdisc == none )
	{
	    strcpy ( TcQdiscAdd , "" );
	    strcpy ( TcQdiscDel , "" );
        }
    }
    if ( UseTcFilter ) 
    {
	for ( unsigned int i = 0; i < tcfilter.size(); i++ )
	{
	    if ( tcfilter[i]->prepareTcFilter() == -1 ) return -1;
	}
    }

    return 1;
}

int TcClass::testParams() 
{
    if (( Strict > 1 ) || ( Strict < 0 )) error->dump( 41, "" );
    
    if ( error->fatalError()) return -1;
    else return 1;
}

int TcClass::envInit()
{
    for (unsigned int i = 0; i < tcfilter.size(); i++) 
    {
	if (tcfilter[i]->envInit() == -1) return -1;
    }
    return 0;
}

unsigned int TcClass::check(std::vector < char * > &iptables_summary, struct timeval now, float cycle_time)
{
    Rate = MeasureRate = 0;
    for ( int i = tcfilter.size()-1 ; i >= 0 ; i--) Rate += tcfilter[i]->check(iptables_summary);
    Rate = MeasureRate = (unsigned int)((float) Rate / cycle_time);
    
    if (Work) {
	if (Rate) Alive = now.tv_sec;
	else if (((now.tv_sec - Alive) >= Hold) && (Hold)) del();
    }
    else if (Rate) { 
	add(); 
	Alive = now.tv_sec; 
    }
    
    if ( SumInSection) {
	if ( Rate > HtbCeil ) Rate = HtbCeil;
	return Rate;
    }
    else return 0;
}

/*unsigned int TcClass::expect_rate( unsigned int expect_rate )
{
    if ( sum_in_section ) 
    { 
	if ( rate < expect_rate ) 
	{
	    unsigned int inc = 4;
	    return ( rate + (power(100*rate/htb_ceil,inc) / power(100,inc-1)) * ((expect_rate-htb_ceil)/100) );
	}
	else 
	{
	    return expect_rate; 
	}
    }
    else return 0;
}*/

int TcClass::add()
{
    Work = true;
    HtbRate = 0;
    HtbCeil = NsCeil;
    OldHtbCeil = HtbCeil;
    OldHtbRate = 0;
    if ( SumInSection ) working++;
    if ( UseTcClass ) {
	sys->tc( TcClassAdd );
	if ( Qdisc!=none ) sys->tc( TcQdiscAdd );
    }
    if ( UseTcFilter) for ( unsigned int i=0; i<tcfilter.size(); i++ ) tcfilter[i]->add();
    
    return 1;
}

int TcClass::del()
{
    Work = false;
    if ( UseTcFilter ) for ( unsigned int i = 0; i < tcfilter.size(); i++ ) tcfilter[i]->del();
    
    if ( SumInSection ) working--;
    if ( UseTcClass ) {
	if ( Qdisc != none ) sys->tc( TcQdiscDel );
	sys->tc( TcClassDel );
    }
    return 1;
}

int TcClass::computeTcChanges()
{
    if ( !Work ) return 0;
    if ( !SumInSection ) return 0;
    if ( HtbCeil > NsCeil ) HtbCeil = NsCeil;
    if ( HtbCeil < NsLow ) HtbCeil = NsLow;

    HtbRate = SectionShape / working; 
    
    if ( HtbCeil < HtbRate ) HtbRate = HtbCeil;

    if (( OldHtbCeil == HtbCeil ) && ( OldHtbRate == HtbRate )) return 0;

    OldHtbRate = HtbRate;
    OldHtbCeil = HtbCeil;
    
    return 1;
}

int TcClass::applyChanges()
{
    if ( !computeTcChanges()) return 0;

    gen_fixed_tc_rateceil_buf ( MsBuf,  HtbRate, HtbCeil ); 

    sprintf ( MlBuf, "tc class change dev %s classid 1:%d htb %s prio %d quantum %d\n", Dev, ClassId, MsBuf, HtbPrio, compute_quantum(HtbCeil));
    
    sys->tc (MlBuf);
    
    return 1;
}

int TcClass::writeChanges(int fd)
{
    char buf[LONG_BUF];

    if ( !computeTcChanges()) return 0;

    sprintf( buf, "class change dev %s classid 1:%d htb rate %d ceil %d prio %d quantum %d\n", Dev, ClassId, HtbRate, HtbCeil, HtbPrio, compute_quantum(HtbCeil));

    write( fd, buf, strlen(buf));    
    
    return 1;
}

void TcClass::stats(int pipe_write, statsclassestype stats_show)
{
    char buf[STAT_MESSAGE_BUF];
  
    if (( stats_show == c_all ) || 
	(( stats_show == c_active ) && MeasureRate ) ||
	(( stats_show == c_working ) && Work ))
    {
	//
    }	
    else return;
  
    if ( StatsShowHtbCeil && StatsShowRate ) sprintf( buf, "%s %d %d\n", Name, HtbCeil, MeasureRate );
    else if ( StatsShowHtbCeil ) sprintf( buf, "%s %d -\n", Name, HtbCeil );
    else if ( StatsShowRate ) sprintf( buf, "%s - %d\n", Name, MeasureRate );
    
    char _buf_ = PROTO_BASE + strlen(buf);
    
    write( pipe_write, &_buf_, 1 );
    write( pipe_write, buf, strlen(buf));
}
