/*                                                                                                                                                                                 
 *      NiceShaper - Dynamic Traffic Management                                                                                                                                    
 *                                                                                                                                                                                 
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>                                                                                                              
 *                                                                                                                                                                                 
 *      This file is subject to the terms and conditions of the GNU General Public                                                                                                 
 *      License.  See the file COPYING in the main directory of this archive for                                                                                                   
 *      more details.                                                                                                                                                              
 */   

#include "conv.h"

Conv::Conv ()
{
    include = new Include(); 
}

Conv::~Conv ()
{
    delete include;
}

int Conv::convToFpv ( vector < char * > &fpv, FILE * fp , filetype type, vector < unsigned int > &protected_marks, bool add_to_protected_marks, char sections [])
{
    int cbuf;
    int buf_counter = 0;    
    bool state_store = true, potential_block_comment_tag = false, state_block_comment = false;
    char * buf = new char [LONG_BUF];
    char sec[SHORT_BUF]="";
    bool running_class=false;
    int n;
    bool file_defined=false;
    FILE *incfp;
    
    vector < char * > fpv_tmp;

    while ((cbuf = fgetc(fp)) != -1 )
    {
        // '<'-60, '#'-35, '>'-62
        if ( !state_block_comment ) {
	    if (potential_block_comment_tag) {
    	        if ( (cbuf == 35) && state_store ) {
		    state_block_comment = true;
		    buf_counter--;
		    buf[buf_counter]=0;
		    continue;
		}
	        potential_block_comment_tag = false;
	    }
	    if ( cbuf == 60 ) potential_block_comment_tag = true; 		
	}
	else if ( state_block_comment ) {
	    if ( potential_block_comment_tag ) {
		if ( cbuf == 62 ) {
		    state_block_comment = false;
		    state_store = true;
	    	    buf[buf_counter]=0;
	    	    continue;
		}
		potential_block_comment_tag = false;
	    }
	    if ( cbuf == 35 ) potential_block_comment_tag = true;
	}

        if ( cbuf == 35 ) state_store = false;	// '#'
	if ( cbuf == 59 ) cbuf = 10;		// ';'
	if ( cbuf == 10 ) state_store = true;	// 'NewLine'
		
	if ( state_store && !state_block_comment)
	{
	    buf[buf_counter]=cbuf;
	    buf_counter++;
	    if ( buf_counter == LONG_BUF-1 ) { error->dump( 25, buf ); return -1; }
	}
	
	if ( cbuf != 10 ) continue; 
	
	buf[buf_counter]=0;
	buf_counter=0;
	state_store = true;	    	    
	
	if ( trim(buf)==-1 ) { return -1; }

	if ( test->validWhiteLine(buf)) continue;
        
        if ( !strcmp( awk( buf, 1 ), "include" )) {
	    if ( fpv_tmp.size()) {
	        for ( unsigned int i = 0; i < fpv_tmp.size(); i++ ) {
    	            if (running_class) fpv.push_back ( fpv_tmp[i] );
	        }
    	        fpv_tmp.clear();
	    }
    		
	    if ( include->fill( buf, file_defined ) == -1 ) return -1;		
	    if ( !file_defined ) continue;
	    if ( !test->fileIsReadable( include->Path )) { error->dump( 26, buf ); return -1; }
	    
	    incfp = fopen( include->Path, "r" );
	    fseek ( incfp, 0, SEEK_SET );
	    
	    if ( !strcmp( include->Type, "file" )) {
	        if ( convToFpv( fpv, incfp , type , protected_marks, add_to_protected_marks , sections ) == -1 ) { fclose (incfp); return -1; }
	    }
    	    else if ( !strcmp( include->Type, "usersfile" )) {
	        if ( !strlen( include->IfaceInet )) {
		    strcpy(include->IfaceInet, "eth0");
		    warning->dump( 5, "iface-inet eth0" );
		}
		if ( !strlen( include->DownloadSection )) {
		    warning->dump( 6, "download-section" );
		}	
		if ( !strlen( include->UploadSection )) {
		    warning->dump( 6, "upload-section" );
		}	
		if ( convertUsersFile ( fpv, incfp , include->IfaceInet, include->DownloadSection, include->UploadSection, sections, include->ResolveHostName, false ) == -1 ) { fclose (incfp); return -1; }
		for ( unsigned int i = 0; i < fpv.size(); i++ ) {
		    if ( trim(fpv[i])==-1 ) { return -1; }
		}
	    }
	    else {
	        error->dump( 24, buf );
		return -1;
	    }
	    fclose (incfp); 	
	    file_defined=false;
	    continue;
	}
	if ( type==classtype ) {
	    if ( !strcmp( awk( buf, 1 ), "class" )) {
	        if ( fpv_tmp.size()) {
    	    	    for ( unsigned int i = 0; i < fpv_tmp.size(); i++ ) {
		        if (running_class) fpv.push_back(fpv_tmp[i]);
		    }
		    fpv_tmp.clear();
		}
		n=1;                                                                                                                               
        	running_class=false;                                                                                                                       
		strcpy ( sec, awk( buf, 2 ));
	    	while ( strlen( awk( sections, n ) )) {
        	    if ( !strcmp ( awk( sections, n ), sec )) {
                	running_class = true; break;
        	    }
            	    n++;
    		}
	    }
	    if (( !strcmp( awk( buf, 1 ), "match" )) || ( !strcmp( awk2( buf, 1 ), "class" ))) { 
	        if ( running_class ) fpv.push_back (buf);
	    }
  	    else {
	        if ( running_class ) {
		    if ( directiveSplit ( fpv_tmp, buf ) == -1 ) return -1;
		}
	    }
	}
	else {
	    if ( directiveSplit ( fpv, buf ) == -1 ) return -1;
	}

    	if ( !add_to_protected_marks ) continue;
	    
        if (( type==conftype ) && ( !strcmp( awk( buf, 1 ), "iface" ))) {
            if ( strstr (buf, " mark " ) ) protected_marks.push_back( atoi( awk( strstr ( buf, " mark "), 2 )));
        }
        else if (( type==classtype ) && ( !strcmp( awk( buf, 1 ), "match" ))) {
    	    if ( strstr( buf, " mark " )) protected_marks.push_back( atoi( awk( strstr ( buf, " mark " ), 2 )));	
            if ( strstr( buf, " set-mark " )) protected_marks.push_back( atoi( awk( strstr (buf, " set-mark "), 2 )));			
        }
        else if (( type==classtype ) && ( !strcmp( awk( buf, 1 ), "set-mark" ))) protected_marks.push_back( atoi( awk( buf, 2 )));	

	buf = new char [LONG_BUF];			    
    }
    
    if ( ( type==classtype ) && ( fpv_tmp.size()))
    {
        for ( unsigned int i = 0; i < fpv_tmp.size(); i++ ) 
	{
	    if (running_class) fpv.push_back ( fpv_tmp[i] ); 
	}    
        fpv_tmp.clear();			    
    }
    return 0;
}

int Conv::directiveSplit ( vector < char * > &fpv, char arg [] )
{
    char *buf;
    char option[SHORT_BUF], directive[SHORT_BUF], param[SHORT_BUF], value[SHORT_BUF];
    bool nextt1 = false;	    
    unsigned int i = 0, j = 0, n = 0;
    
    // type1 directives. 
    // has 1 required value and nothing else, 
    // if there is another value beyound, it have to be another type1 directive, if it's not, we have syntax error 
    static char t1[10][SHORT_BUF]= { "ceil", "hold", "lang", "low", "mode", "rate", "reload", "set-mark", "strict", "" };
    // type2 directives.
    // gets all given values as his own. Need at least 1 value.
    static char t2[4][SHORT_BUF] = { "debug", "mark-on-ifaces", "run", "" };
    // type3 directives
    // there shouldn't be any values, if there are, we have syntax error     
    static char t3[4][SHORT_BUF]= { "do-not-shape", "wrapper", "virtual", "" };
    // type4 directives
    // by iteration, gets pairs of words ( parameter and value ), 
    // it's syntax error if parameter is unknown or one of pair elements is empty.
    static char t4[10][9][SHORT_BUF] = {{ "include", "file", "" },
				{ "stats", "classes", "file", "group", "mode", "owner", "sum", "unit", "" },
				{ "log", "file", "syslog", "terminal", "" },
				{ "section", "shape", "speed", "" },
				{ "htb", "scheduler", "prio", "" },
				{ "sfq", "perturb", "" },
				{ "esfq", "hash", "perturb", "" },
				{ "iptables", "hook", "append-hook", "insert-hook", "target", "" },
				{ "imq", "autoredirect", "" },
				{ "" }};
    // type5 directives
    // special directives, copy whole line without changes.				
    static char t5[4][SHORT_BUF] = { "class", "iface", "match", "" };
    
    n = 1;
    if ( !strncmp( awk( arg, 1 ), "default", 7 )) n++;
    strncpy ( option, awk( arg, n ), SHORT_BUF );
    
    i = 0; 
    while ( strlen( t1[i] )) 
    {
	if ( !strcmp( t1[i], option )) 
	{
	    strcpy( directive, option );
	    strncpy( value, awk( arg, ++n ), SHORT_BUF );
	    if ( !strlen(value) )
	    {
		error->dump( 24, arg );
	        return -1;
	    }

	    buf = new char [LONG_BUF+1];
	    strncpy( buf, "", LONG_BUF+1 ); 
	    sprintf( buf, "%s %s", directive, value );
	    fpv.push_back (buf);
	    if ( strlen(awk( arg, ++n ))) 
	    {
	        strncpy ( option, awk( arg, n ), SHORT_BUF );
	        j = 0; nextt1 = false;
		while ( strlen(t1[j] )) 
		{
		    if ( !strcmp( t1[j], option )) 
		    {
			nextt1 = true;
			break;
		    }
		    else j++;
		}
		if ( !nextt1 )
		{
		    error->dump( 24, arg );
	    	    return -1;
	        }
	    }
	    if ( !nextt1 ) return 1;
	}
	i++;
	if ( nextt1 ) 
	{
	    i = 0;
	    nextt1 = false;
	}
    }
    
    i = 0;
    while ( strlen( t2[i] )) 
    {
        if ( !strcmp( t2[i], option ))
	{
	    strcpy( directive, option );
	
	    while ( strlen( awk( arg, ++n ))) 
    	    {
    		strncpy( value, awk( arg, n ), SHORT_BUF );
    		buf = new char [LONG_BUF+1];
		strncpy( buf, "", LONG_BUF+1 );
		sprintf( buf, "%s %s", directive, value );
	        fpv.push_back( buf );

	    }
	    
	    if ( n == 2 )
	    {
		error->dump( 24, arg );
	    	return -1;
	    }
	    else return 1;
        }
	i++;
    }

    i = 0;
    while ( strlen( t3[i] )) 
    {
        if ( !strcmp( t3[i], option )) 
        {
            strcpy( directive, option );
    	    buf = new char [LONG_BUF+1];
    	    strncpy( buf, "", LONG_BUF+1 );
	    sprintf( buf, "%s", directive );
	    fpv.push_back( buf );
	    if ( strlen( awk( arg, ++n ))) 
	    {
		error->dump( 24, arg );
	        return -1;
	    }
	    return 1;
        }
	i++;
    }

    i = 0;
    while ( strlen( t4[i][0] )) 
    {
	if ( !strcmp( t4[i][0], option ))
	{
	    while ( strlen( awk( arg, ++n ))) 
    	    {
	        strcpy( directive, option );
    		strncpy( param, awk( arg, n ), SHORT_BUF );
    	        strncpy( value, awk( arg, ++n ), SHORT_BUF );
    		if ( !strlen(param) || !strlen(value) )
	        {
		    error->dump( 24, arg );
	    	    return -1;
		}

	        j = 1;
		while ( true) 
	        {
		    if ( !strcmp( t4[i][j], param )) break;
		    j++;
		    if ( !strlen( t4[i][j] )) 
		    {
			error->dump( 24, arg );
	    	        return -1;
	    	    }
	        }
	        buf = new char [LONG_BUF+1];
	        strncpy( buf, "", LONG_BUF+1 );
	        sprintf( buf, "%s %s %s", directive, param, value );
		fpv.push_back( buf );	

	    }
	    return 1;
	}
        i++;
    }

    i = 0;
    while ( strlen( t5[i] )) 
    {
        if ( !strcmp( t5[i], option ))
	{
    	    buf = new char [LONG_BUF+1];
    	    strncpy( buf, "", LONG_BUF+1 ); 
	    strncpy( buf, arg, LONG_BUF );
	    fpv.push_back( buf );
	
	    return 1;
        }
	i++;   
    }
    
    
    if ( arg[0] == '<' )
    {
    	buf = new char [LONG_BUF+1];
    	strncpy( buf, "", LONG_BUF+1 ); 
	strncpy( buf, arg, LONG_BUF );
	fpv.push_back( buf );
	
	return 1;	
    }
    
    error->dump ( 11, arg );
    return -1;
}


int Conv::convertUsersFile ( vector < char * > &fpv, FILE * fp, char dev_inet[MAX_IFACE], char section_download[SHORT_BUF], char section_upload[SHORT_BUF], char sections [LONG_BUF], bool resolve_hostname, bool make_comments )
{
    char *buf, usersbuf[LONG_BUF], hostname[MAX_NAME+1], ip[SHORT_IP_ADDR+1], dev[MAX_IFACE+1];

    while ( fgets( usersbuf, LONG_BUF, fp )) 
    {
	if ( !test->validIp( awk( usersbuf , 1 ))) continue;
	if ( !test->validSysDev( awk( usersbuf , 2 ))) continue;
        
        strncpy ( ip, "", SHORT_IP_ADDR+1 ); 
        strncpy ( dev, "", MAX_IFACE+1 ); 

	strncpy( ip, awk( usersbuf , 1 ), SHORT_IP_ADDR );
	strncpy( dev, awk( usersbuf , 2 ), MAX_IFACE );

	if ( !test->validIp(ip)) { error->dump( 29, usersbuf ); return -1; }   
	if ( !test->validSysDev(dev)) { error->dump( 16, usersbuf ); return -1; }   

        if ( resolve_hostname ) iptohostname( hostname , ip);
        else strncpy( hostname, ip, MAX_NAME+1 );

        if ( make_comments )
        {
	    buf=new char[SHORT_BUF];
            sprintf( buf, "\n############### %s ###", hostname );
    	    fpv.push_back(buf);
	}
	
	if ( strlen( section_download )) {
	    buf=new char[SHORT_BUF];
	    sprintf( buf , "class %s %s %s", section_download, dev , hostname );
    	    fpv.push_back(buf);
	    buf=new char[SHORT_BUF];
	    sprintf( buf , "\tmatch dstip %s", ip);
	    fpv.push_back(buf);
    	    if ( strstr ( usersbuf , " dl_low " )) {
		buf=new char[SHORT_BUF];
		strcpy( buf , "\tlow " );
    		strcat( buf , awk ( strstr( usersbuf , " dl_low " ) , 2 ));
		fpv.push_back(buf);
    	    }
    	    if ( strstr ( usersbuf , " dl_ceil " )) {
		buf=new char[SHORT_BUF];
    		strcpy( buf , "\tceil " );
    		strcat( buf , awk ( strstr( usersbuf , " dl_ceil " ) , 2 ));
    		fpv.push_back(buf);
	    }
    	    if ( strstr ( usersbuf , " dl_rate " )) {
		buf=new char[SHORT_BUF];
    	        strcpy( buf , "\trate " );
    		strcat( buf , awk ( strstr( usersbuf , " dl_rate " ) , 2 ));
		fpv.push_back(buf);
    	    }
	}
	
	if ( strlen( section_upload )) {	
	    buf=new char[SHORT_BUF];
	    sprintf( buf , "class %s %s %s", section_upload, dev_inet , hostname );
	    fpv.push_back(buf);
	    buf=new char[SHORT_BUF];
    	    sprintf( buf , "\tmatch srcip %s", ip);
	    fpv.push_back(buf);
    	    if ( strstr ( usersbuf , " ul_low " )) {
		buf=new char[SHORT_BUF];
    	        strcpy( buf , "\tlow " );
    		strcat( buf , awk ( strstr( usersbuf , " ul_low " ) , 2 ));
		fpv.push_back(buf);
    	    }
    	    if ( strstr ( usersbuf , " ul_ceil " )) {
		buf=new char[SHORT_BUF];
    		strcpy( buf , "\tceil " );
    		strcat( buf , awk ( strstr( usersbuf , " ul_ceil " ) , 2 ));
		fpv.push_back(buf);
    	    }
    	    if ( strstr ( usersbuf , " ul_rate " )) {
		buf=new char[SHORT_BUF];
    		strcpy( buf , "\trate " );
		strcat( buf , awk ( strstr( usersbuf , " ul_rate " ) , 2 ));
		fpv.push_back(buf);
    	    }
	}    
    }

    return 0;
}

Include::Include ()
{
    strcpy ( Path, "" );
    strcpy ( Type, "" );
    strcpy ( DownloadSection, "" );
    strcpy ( UploadSection, "" );
    strcpy ( IfaceInet, "" );
    ResolveHostName=false;
}

int Include::fill ( char arg [], bool &file_defined )
{
    char param[SHORT_BUF], value[LONG_BUF];
    unsigned int n=1;
    
    while ( strlen( awk( arg, ++n ))) {
        strcpy( param, awk( arg, n ));
        strcpy( value, awk( arg, ++n ));
	
	if ( !strlen(param)) { error->dump( 24, arg ); return -1; }
	if ( !strlen(value)) { error->dump( 24, arg ); return -1; }
	
	if ( !strcmp( param, "file" )) {
	    strcpy( Path, value );
	    strcpy( Type, "file" );
	    file_defined=true;
	}
	else if ( !strcmp( param, "usersfile" )) {
	    strcpy( Path, value );
	    strcpy( Type, "usersfile" );
	    file_defined=true;
	}
	else if ( !strcmp( param, "download-section" )) {
	    strcpy( DownloadSection, value );	
	}
	else if ( !strcmp( param, "upload-section" )) {
	    strcpy( UploadSection, value );		
	}
	else if ( !strcmp( param, "iface-inet" )) {
	    strcpy( IfaceInet, value );		
	}
	else if ( !strcmp( param, "resolve-hostname" )) {
	    if ( !strcmp( value, "true" )) ResolveHostName=true;
	    else if ( !strcmp( value, "false" )) ResolveHostName=false;
	    else { error->dump( 24, arg ); return -1; }
	}
	else { 
	    error->dump( 24, arg );
	    return -1;
	}
    }
    
    return 1;
}

