/*                                                                                                                                                                                 
 *      NiceShaper - Dynamic Traffic Management                                                                                                                                    
 *                                                                                                                                                                                 
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>                                                                                                              
 *                                                                                                                                                                                 
 *      This file is subject to the terms and conditions of the GNU General Public                                                                                                 
 *      License.  See the file COPYING in the main directory of this archive for                                                                                                   
 *      more details.                                                                                                                                                              
 */   

#include "filter.h"

TcFilter::TcFilter( char buf[], char dev[], filtertype filter )
{
    FilterId = 0 ;
    FlowId = 0;
    HandleFw = 0;
    Chains = 0;
    ImqAutoRedirect = true;
    DestMark = false;
    FromLocalhost = false;
    ToLocalhost = false;
    UseTcFilter = true;
    UseForFw = false;
    Filter = u32;
    IptTarget = t_accept;
    strcpy( Dev, dev );
    Filter = filter;
    strcpy( Match, &(buf)[5] );
    trim( Match );
    if ( strstr( Match, " use-for-fw true" )) UseForFw = true;
    if ( strstr( Match, " set-mark " )) HandleFw = read_fw_mark( awk( strstr( Match, " set-mark " ), 2 ));
    if ( strstr( Match, " filterid " )) FilterId = atoi( awk( strstr( Match, " filterid " ), 2 ));
    if ( strstr( buf, "to localhost" )) ToLocalhost = true;
    else if ( strstr( buf, "from localhost" )) FromLocalhost = true;
}

TcFilter::~TcFilter()
{
    char buf[LONG_BUF];

    if ( FromLocalhost ) 
    {
	sprintf( buf, "iptables -t mangle -D OUTPUT %s -j %s ", IptMatch, ipt_mychain ); 
	sys->ipt( buf ); 
    }
    if ( ToLocalhost ) 
    {
	sprintf( buf, "iptables -t mangle -D INPUT %s -j %s ", IptMatch, ipt_mychain ); 
	sys->ipt( buf ); 
    }
}
int TcFilter::store(char buf[])
{
    char option[SHORT_BUF], param[SHORT_BUF], value[SHORT_BUF];

    strcpy( option, awk( buf, 1 ));
    strcpy( param, awk( buf, 2 ));
    strcpy( value, awk( buf, 3 ));
    if ( !strcmp( option, "class" )) 
    {
	//
    }
    else if ( !strcmp( option, "classid" )) FlowId = atoi(param);
    else if ( !strcmp( option, "standard-class" )) UseTcFilter = true;      
    else if ( !strcmp( option, "do-not-shape" )) { UseTcFilter = true; FlowId = 1; } 
    else if ( !strcmp( option, "virtual" )) { UseTcFilter = false; IptTarget = t_empty; } 
    else if ( !strcmp( option, "wrapper" )) UseTcFilter = true;
    else if ( !strcmp( option, "imq" )) {
	if ( !strcmp( param, "autoredirect" )) {
    	    if ( !strcmp( value, "true" )) ImqAutoRedirect = true;
    	    else if ( !strcmp( value, "false" )) ImqAutoRedirect = false;
	    else { error->dump( 11, buf ); }
	}
	else { error->dump( 11, buf ); }
    }
    else if ( !strcmp( option, "iptables" ))
    {
	if ( !strcmp( param, "target" )) 
	{
	    if ( !strcmp( value, "accept" )) IptTarget = t_accept;
	    else if ( !strcmp( value, "return" )) IptTarget = t_return;
	    else if ( !strcmp( value, "drop" )) IptTarget = t_drop;
	    else if ( !strcmp( value, "empty" )) IptTarget = t_empty;
	    else { 
		error->dump( 11, buf ); 
		return -1; 
	    }
	}
	else {
	    error->dump( 11, buf );
	    return -1;
	}
    }
  
    return 1;
}

int TcFilter::prepareTcFilter()
{
    char param[SHORT_BUF], value[SHORT_BUF];
    char buf[LONG_BUF] = "";
    char addr[16], mask[16];
    unsigned int n=0;

    strcpy ( TcFilterAdd, "" );
    strcpy ( TcFilterDel, "" );
    if ( !UseTcFilter ) return 0;
    if ( Filter == u32 ) {
      	sprintf( TcFilterAdd, "tc filter add dev %s protocol ip parent 1:0 prio 10 handle 800::%x u32 ", Dev, FilterId );
	sprintf( TcFilterDel, "tc filter del dev %s protocol ip parent 1:0 prio 10 handle 800::%x u32", Dev, FilterId );

	while ( strlen( awk( Match, ++n ))) {
	    strcpy ( param, awk( Match, n ));
	    strcpy ( value, awk( Match, ++n ));
	    
	    if ( !strcmp( param, " proto " )) {
		if ( !strcmp( value, "tcp" )) strcat( TcFilterAdd, " match ip protocol 6 0xff " );
	        else if ( !strcmp( value, "udp" )) strcat( TcFilterAdd, " match ip protocol 17 0xff " );
		else if ( !strcmp( value, "icmp" )) strcat( TcFilterAdd, " match ip protocol 1 0xff " );
    	    }
	    else if ( !strcmp( param, "srcip" )) {
		if ( split( addr, mask, value ) == -1 ) return -1;
    	        if ( test->solidIpMask( mask )) sprintf( buf, " match ip src %s/%d ", addr, dotobit( mask ));
	        else { error->dump( 37, Match ); return -1; }
	        strcat( TcFilterAdd, buf );
	    }
	    else if ( !strcmp( param, "dstip" )) {
	        if ( split( addr, mask, value ) == -1 ) return -1;
    	        if ( test->solidIpMask( mask )) sprintf( buf, " match ip dst %s/%d ", addr, dotobit( mask ));
	        else { error->dump( 37, Match ); return -1; }
	        strcat( TcFilterAdd, buf );
	    }
    	    else if (( !strcmp( param, "srcport" )) || ( !strcmp( param, "sport" ))) {
    		strcat( TcFilterAdd, " match ip sport " );
		strcat( TcFilterAdd, value );
		strcat( TcFilterAdd, " 0xffff " );
	    }
    	    else if (( !strcmp( param, "dstport" )) || ( !strcmp( param, "dport" ))) {
    		strcat( TcFilterAdd, " match ip dport " );
		strcat( TcFilterAdd, value );
		strcat( TcFilterAdd, " 0xffff " );
	    }
	}
      	sprintf( buf, " flowid 1:%d", FlowId );
	strcat( TcFilterAdd, buf );  
    } 
    else if ( Filter == fw ) { 
	if ( !HandleFw ) HandleFw = FilterId;
	sprintf( TcFilterAdd, "tc filter add dev %s protocol ip parent 1:0 prio 10 handle %d fw flowid 1:%d", Dev, HandleFw, FlowId );
	sprintf( TcFilterDel, "tc filter del dev %s protocol ip parent 1:0 prio 10 handle %d fw", Dev, HandleFw );
    }
    return 1;
}
int TcFilter::envInit()
{
    char buf[LONG_BUF] = "";
    bool need_mark;

    if ( gen_ipt_filter ( need_mark, IptMatch, Match ) == -1 ) return -1;

    if (( need_mark ) && ( Filter != fw )) {
	error->dump( 36, Match ); 
	return -1;     
    }

    if (FromLocalhost) {
	strcat(IptMatch, " -o ");
	strcat(IptMatch, Dev);
	strcat(IptMatch, " ");
	sprintf(buf, "iptables -t mangle -A OUTPUT %s -j %s ", IptMatch, ipt_mychain);
	sys->ipt(buf);
    }
    if (ToLocalhost) {
	strcat(IptMatch, " -i ");
	strcat(IptMatch, Dev);
	strcat(IptMatch, " ");
	sprintf(buf, "iptables -t mangle -A INPUT %s -j %s ", IptMatch, ipt_mychain);
	sys->ipt(buf);
    }
    if (Filter == fw) {
	Chains++;
	sprintf(buf, "iptables -t mangle -A %s %s -j MARK --set-mark %d", ipt_mychain, IptMatch, HandleFw);
	sys->ipt(buf);
    }
    if (strstr(Dev, "imq")) {
	if (ImqAutoRedirect) {
	    Chains++;
	    sprintf(buf, "iptables -t mangle -A %s %s -j IMQ --todev %s", ipt_mychain, IptMatch, &(Dev[3]));
	    sys->ipt(buf);
	}
    }
    Chains++;  
    if ( IptTarget == t_drop ) sprintf(buf, "iptables -t mangle -A %s %s -j DROP", ipt_mychain, IptMatch);
    else if ( IptTarget == t_return ) sprintf(buf, "iptables -t mangle -A %s %s -j RETURN", ipt_mychain, IptMatch);
    else if ( IptTarget == t_empty ) sprintf(buf, "iptables -t mangle -A %s %s", ipt_mychain, IptMatch);
    else sprintf(buf, "iptables -t mangle -A %s %s -j ACCEPT", ipt_mychain, IptMatch);
    sys->ipt(buf);
      
    return 1;
}

unsigned int TcFilter::check(std::vector < char * > &iptables_summary)
{
    register unsigned int n, m;
    static unsigned int state, value;    
    static char buf[LONG_BUF];

    if ( iptables_summary.size() < Chains ) {
	error->dump( 12, "" );
        return 0;
    }

    for ( n=0; n<(Chains-1); n++ ) { 
	delete [] iptables_summary.back();
	iptables_summary.pop_back();
    }
    
    strncpy ( buf, iptables_summary.back() , SHORT_BUF );
    delete [] iptables_summary.back();
    iptables_summary.pop_back();

    state=value=0;
    for ( n=0; n<SHORT_BUF ; n++ ) {
	m=buf[n];
	if (( m>=65 && m<=90 ) || ( m>=97 && m<=122 )) { state=n; break; } 
    }
    if ( !state ) { error->dump( 12, "" ); return 0; }

    for ( n=state; n>0; n-- ) {
	m=buf[n];
	if ( m>=48 && m<=57 ) { state=n; break;	}
    }
    if (!n) { error->dump( 12, "" ); return 0; }

    unsigned int mul=1;
    for ( n=state; n>0; n-- ) {
	m=buf[n];
	if ( m>=48 && m<=57 ) {
	    m-=48;
	    value+=mul*m;
	    mul*=10;
	}
	else break;
    }

    return value << 3;
}

unsigned int TcFilter::add()
{
    if ( Filter == u32 ) sys->tc(TcFilterAdd);
    else if (( Filter == fw ) && ( UseForFw )) sys->tc(TcFilterAdd);
    else return 0;
    
    return 1;
}

unsigned int TcFilter::del()
{
    if ( Filter == u32 ) sys->tc(TcFilterDel);
    else if (( Filter == fw ) && ( UseForFw )) sys->tc(TcFilterDel);
    else return 0;
    
    return 1;
}

