/*                                                                                                                                                                                 
 *      NiceShaper - Dynamic Traffic Management                                                                                                                                    
 *                                                                                                                                                                                 
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>                                                                                                              
 *                                                                                                                                                                                 
 *      This file is subject to the terms and conditions of the GNU General Public                                                                                                 
 *      License.  See the file COPYING in the main directory of this archive for                                                                                                   
 *      more details.                                                                                                                                                              
 */   

#include "instance.h"
#include "errno.h"

Instance::Instance( char stats_unit_str[], unittype stats_unit )
{
    int fd1[2], fd2[2];

    pipe(fd1);
    pipe(fd2);
    
    if ((Pid = fork())) {
	i_am_parent = true;
	close(fd1[1]);
	close(fd2[0]);
	PipeRead = fd1[0];
	PipeWrite = fd2[1];
	fcntl(PipeRead, F_SETFL, 0);
	fcntl(PipeWrite, F_SETFL, 0);
	fcntl(fd1[0], F_SETFL, 0);
	fcntl(fd2[1], F_SETFL, 0);
	strncpy( StatsUnitStr, "", SHORT_BUF+1 ); 
	strncpy( StatsUnitStr, stats_unit_str, SHORT_BUF ); 
	statsunit = stats_unit;
    } 
    else {
	i_am_parent = false;
	Pid = getpid();
	close(fd2[1]);
	close(fd1[0]);
	PipeRead = fd2[0];
	PipeWrite = fd1[1];
	fcntl(PipeRead, F_SETFL, 0);
	fcntl(PipeWrite, F_SETFL, 0);
	fcntl(fd1[1], F_SETFL, 0);
	fcntl(fd2[0], F_SETFL, 0);
    }
}

Instance::~Instance()
{
    close(PipeRead);
    close(PipeWrite);    
}

void Instance::killChild()
{
    kill(Pid, SIGTERM);
    waitpid(Pid, NULL, 0);
}

void Instance::sendInitStatusOk()
{
    char buf = ACK;

    write(PipeWrite, &buf, 1);
}

void Instance::sendInitStatusError()
{
    char buf = NAK;

    write(PipeWrite, &buf, 1);
}

int Instance::getInitStatus()
{
    char buf;

    if ( read(PipeRead, &buf, 1 ) != 1 ) return -1;
    
    if ( buf==(ACK)) 
	return 0;
    else 
	return -1;
}

void Instance::sendContinue()
{
    char _buf_=EOT;
    
    write( PipeWrite, &_buf_, 1 );
}

void Instance::sendResult( NiceShaper *ns )
{
    char _buf_=EOT;

    ns->stats(PipeWrite);

    write( PipeWrite, &_buf_, 1 );
}

int Instance::sendFatalError()
{
    char _buf_=NAK;
    
    write( PipeWrite, &_buf_, 1 );
    
    return 0;
}

int Instance::waitForContinue()
{
    char _buf_ = 0 ; 

    if ( read( PipeRead, &_buf_ , 1 ) == -1 ) return -1; 
    
    if ( _buf_ == (EOT)) return 0;
    else return -1;
}

int Instance::getResult()
{
    char _buf_ = 0 ; 
    char *buf;
    int msg_len, got_msg_len;

    signal( SIGUSR1, signalstats );
    if ( read( PipeRead, &_buf_ , 1 ) != 1 ) { error->dump(42,""); return -1; }
    signal( SIGUSR1, SIG_IGN );

    if ( _buf_ == (NAK)) return -1; 
    else if ( _buf_ == (EOT)) return 0;
    else msg_len=_buf_-PROTO_BASE;

    for ( int n=StatsContainer.size()-1; n>=0; n-- )
    {
	delete [] StatsContainer.back();
	StatsContainer.pop_back();    
    }
    
    do {
        if (( msg_len < 0 ) || ( (unsigned int)msg_len >= STAT_MESSAGE_BUF )) { error->dump( 42, "" ); return -1; }
        got_msg_len = 0;
        buf = new char[STAT_MESSAGE_BUF];
        buf[msg_len]=0;
        while (msg_len) 
        {
            if (( got_msg_len=read( PipeRead, buf + got_msg_len, msg_len )) <= 0 ) return -1;
	    msg_len -= got_msg_len;
	    usleep (10000);
        }
	StatsContainer.push_back (buf);	
	if ( read( PipeRead, &_buf_ , 1 ) != 1 ) { error->dump( 42, "" ); return -1; }
        msg_len=_buf_-PROTO_BASE;
    } while ( _buf_ != (EOT));

    return 0; 
}

void Instance::dumpStats(int fd)
{
    char buf[STAT_MESSAGE_BUF], buf2[STAT_MESSAGE_BUF], *stat_line;

    for ( unsigned int n=0; n<StatsContainer.size(); n++ ) {
	stat_line = StatsContainer[n];
	if (n) {
	    sprintf( buf, "%-32.32s  ", awk( stat_line, 1 ));
	    strcpy( buf2, "" );
	    if ( strcmp( awk( stat_line, 2 ), "-" )) {
		sprintf( buf2, "%8d%s - ", (int) unitconvert( atoi( awk( stat_line, 2 )), statsunit ), StatsUnitStr );
	    }
	    else {
    		for ( unsigned int i=0 ; i < 8+strlen (StatsUnitStr) ; i++ ) strcat (buf2," ");
		strcat (buf2, "   ");
	    }
	    strcat( buf, buf2 );
	    strcpy( buf2, "" );
	    if ( strcmp( awk(stat_line, 3), "-" ) ) {
		sprintf(buf2, "( %8d%s )\n", (int) unitconvert( atoi( awk( stat_line, 3 )), statsunit ),  StatsUnitStr );
    	    }
	    else {
		strcat( buf2, "( " );
    		for ( unsigned int i=0 ; i < 8+strlen ( StatsUnitStr ) ; i++ ) strcat ( buf2, " " );
    		strcat( buf2, " )\n" );           	
	    }	
	    strcat( buf, buf2 );
	} 
	else {
	    sprintf( buf, "%-32.32s  ", awk( stat_line, 1 ));
	    for (unsigned int n = 0; n < strlen( StatsUnitStr ); n++) strcat( buf, " " );
	    sprintf( buf2, "%8s - (", awk( stat_line, 2 ));
	    strcat( buf, buf2);
	    for (unsigned int n = 0; n < strlen( StatsUnitStr-5 ); n++) strcat( buf, " " );
	    sprintf( buf2, " %8s )\n", awk( stat_line, 3 ));
	    strcat( buf, buf2 );
	}
	write( fd, buf, strlen(buf));
    }
    
    sprintf(buf, "\n");
    write(fd, buf, 1);
}
