/*                                                                                                                                                                                 
 *      NiceShaper - Dynamic Traffic Management                                                                                                                                    
 *                                                                                                                                                                                 
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>                                                                                                              
 *                                                                                                                                                                                 
 *      This file is subject to the terms and conditions of the GNU General Public                                                                                                 
 *      License.  See the file COPYING in the main directory of this archive for                                                                                                   
 *      more details.                                                                                                                                                              
 */   

#include "main.h"
#include "instance.h"
#include "conv.h"
#include "aux.h"
#include "error.h"
#include "sys.h"
#include "niceshaper.h"

char conffile[MAX_PATH]="/etc/niceshaper0.6/config.ns";
char classfile[MAX_PATH]="/etc/niceshaper0.6/class.ns";
char statsfile[MAX_PATH] = "/tmp/niceshaper.stats";
char statstmp[MAX_PATH] = "/tmp/niceshaper.stats.tmp";
char pidfile[MAX_PATH] = "/var/run/niceshaper.pid";

char section[MAX_NAME+1];
char ipt_mychain[SHORT_BUF];

unsigned int section_id = 0;
unsigned int working = 0;

langtype lang = en;
bool i_am_parent = true; 
bool daemon_mode = true;

class NiceShaper *ns;
class Instance *parent_instance = NULL;
class Sys *sys;
class Tests *test;
class Error *error;
class Warning *warning;
class Info *info;

vector < Instance * > child_instance;

int main(int argc, char *argv[])
{
    char sessions[LONG_BUF] = "", buf[LONG_BUF+1] = "";
    char option[SHORT_BUF+1] = "", param1[SHORT_BUF+1] = "", param2[SHORT_BUF+1] = "";
    char statsfileowner[SHORT_BUF] = "root";
    char statsfilegroup[SHORT_BUF] = "root";
    char statsfilemode[SHORT_BUF] = "644";
    char statspath[MAX_PATH] = "";  
    char statsunitstr[SHORT_BUF] = "kB/s";
    unittype statsunit = kbps;
    bool dump_stats = false;
    bool log_on_terminal = true;
    bool log_to_syslog = true;
    bool log_to_file = false;
    vector < char * > fpv_conffile,fpv_classfile;  
    vector < unsigned int > protected_marks;
    vector < char * >::iterator fpvi, fpvi_begin, fpvi_end;
    struct timeval now, past;
    int fd;
    FILE *fp;
    int n;
    unsigned int reload = 50;
    unsigned int reload_quantum = 100000;
    float frel = 0;
    class Conv *conv;
    
    sys = new Sys;
    error = new Error;
    warning = new Warning;
    info = new Info;
    
    run( argc, argv );

    if ( daemon_mode ) {
	if ( fork()) {
	    while ( access( pidfile, 0 )) usleep(100000);
  	    usleep(100000);
  	    exit(0);
	}
    }

    conv = new Conv;
    test = new Tests;
        
    fp = fopen( conffile, "r" );
    fseek( fp, 0, SEEK_SET );
    if ( conv->convToFpv( fpv_conffile, fp , conftype , protected_marks, true , sessions ) == -1 ) error->markFatalError();
    fclose(fp);
    
    if ( error->fatalError()) start_quit_seq();

    if ( fpv_section_i( fpvi_begin, fpvi_end, fpv_conffile, "global" ) == -1 ) start_quit_seq(); 
    fpvi = fpvi_begin;                                                

    /*while ( fpvi != fpvi_end ) {                                                                        
	printf("cf:%s\n",*fpvi);                                      
	fpvi++;                                                                                    
    }*/

    fpvi=fpvi_begin;
    while ( fpvi != fpvi_end ) {
	strncpy ( buf , *fpvi, LONG_BUF );
	fpvi++;
	strncpy ( option, awk( buf, 1 ), SHORT_BUF );
	strncpy ( param1, awk( buf, 2 ), SHORT_BUF );
	strncpy ( param2, awk( buf, 3 ), SHORT_BUF );
	if ( !strcmp( option, "stats" )) 
	{
  	    if ( !strcmp( param1, "unit" )) 
	    {
  		statsunit = getunit( param2 );
		strcpy( statsunitstr, param2 );
	    }
  	    else if ( !strcmp( param1, "file" )) 
	    {
		strcpy( statspath, param2 );
		if ( !strcmp( statspath, "none" ) || !strcmp( statspath, "false" ) || !strlen( statspath )) dump_stats = false;
		else dump_stats = true;
	    }
	    else if ( !strcmp( param1, "owner" )) {
		strcpy( statsfileowner, param2 );
  	    }
	    else if ( !strcmp( param1, "group" )) {
		strcpy( statsfilegroup, param2 );
	    }
	    else if ( !strcmp( param1, "mode" )) {
		strcpy( statsfilemode, param2 );
	    }
	} 
	else if ( !strcmp( option, "run" )) {
	    strcat( sessions, param1 );
	    strcat( sessions, " " );
	}
	else if ( !strcmp( option, "lang" )) {	
	  if ( !strcmp ( param1 , "en" )) lang=en;	
	  else if ( !strcmp ( param1 , "pl" )) lang=pl;	
	  else lang=en;
	}
	else if ( !strcmp( option, "log" )) {
  	    if ( !strcmp( param1, "terminal" )) {
		if ( !strcmp( param2, "true" )) log_on_terminal = true;
		else if ( !strcmp( param2, "false" )) log_on_terminal = false;
		else error->dump ( 11, buf );
		error->setLogOnTerminal ( log_on_terminal );
		warning->setLogOnTerminal ( log_on_terminal );
		info->setLogOnTerminal ( log_on_terminal );
	    }	
  	    else if ( !strcmp( param1, "syslog" )) {
		if ( !strcmp( param2, "true" )) log_to_syslog = true;
		else if ( !strcmp( param2, "false" )) log_to_syslog = false;
		else error->dump ( 11, buf );
		error->setLogToSyslog ( log_to_syslog );
		warning->setLogToSyslog ( log_to_syslog );
		info->setLogToSyslog ( log_to_syslog );
	    }	
  	    else if ( !strcmp( param1, "file" )) {
		if ( !strcmp( param2, "none" )) log_to_file = false;
		else if ( !strcmp( param2, "false" )) log_to_file = false;
		else {
		    log_to_file = true;

		    error->setLogFile ( param2 );
		    warning->setLogFile ( param2 );
		    info->setLogFile ( param2 );
		}
		error->setLogToFile ( log_to_file );
		warning->setLogToFile ( log_to_file );
		info->setLogToFile ( log_to_file );
	    }	
	}
    }
    
    if ( !strlen ( sessions )) error->dump( 15, "" );
    
    if ( error->fatalError()) start_quit_seq(); 
    
    n = 1;
    while (strlen(awk(sessions, n))) 
    {
        if ( fpv_section_i(fpvi_begin, fpvi_end, fpv_conffile, awk(sessions, n )) == -1 ) start_quit_seq();
        fpvi=fpvi_begin;
    	while ( fpvi != fpvi_end ) {
	    strcpy ( buf , *fpvi );
	    fpvi++;
	    strcpy(option, awk(buf, 1));
	    if ( !strcmp(option, "reload" )) 
	    {
        	reload = (unsigned int)( strtof( awk( buf, 2 ), NULL )*10 );  
            	if (( reload == 0 ) || ( reload > 6000 )) {                                                                                                                                            
		    error->dump ( 13, "" );
		    start_quit_seq();
	        }
	    }
	}
	frel += 1/(float)reload;
	n++;
    }
    reload_quantum=(unsigned int)(100000/frel);
    fp = fopen(classfile, "r");
    fseek(fp, 0, SEEK_SET);
    if ( conv->convToFpv ( fpv_classfile, fp, classtype, protected_marks, true, sessions ) == -1 ) {
	fclose(fp);
	start_quit_seq();
    }
    fclose(fp);
    delete conv;

    test->getSysDevices();

    sys->ipt ("for n in `iptables -t mangle -L PREROUTING -nv --line-numbers | grep ns_ | awk '{print $1}' | sort -r`; do iptables -t mangle -D PREROUTING $n ; done");
    sys->ipt ("for n in `iptables -t mangle -L POSTROUTING -nv --line-numbers | grep ns_ | awk '{print $1}' | sort -r`; do iptables -t mangle -D POSTROUTING $n ; done");
    sys->ipt ("for n in `iptables -t mangle -L INPUT -nv --line-numbers | grep ns_ | awk '{print $1}' | sort -r`; do iptables -t mangle -D INPUT $n ; done");
    sys->ipt ("for n in `iptables -t mangle -L OUTPUT -nv --line-numbers | grep ns_ | awk '{print $1}' | sort -r`; do iptables -t mangle -D OUTPUT $n ; done");
    sys->ipt ("for n in `iptables -t mangle -L FORWARD -nv --line-numbers | grep ns_ | awk '{print $1}' | sort -r`; do iptables -t mangle -D FORWARD $n ; done");
    sys->ipt ("for n in `iptables -t mangle -L -nv | grep 'Chain ns_' | awk '{print $2}'`; do iptables -t mangle -F $n ; iptables -t mangle -X $n;  done");

    n = 1;
    char tag[LONG_BUF];
    char dev[MAX_IFACE];
    char buf2[LONG_BUF];    
  
    while (strlen(awk(sessions, n))) 
    {
	strcpy(tag, awk(sessions, n));  
        if ( fpv_section_i( fpvi_begin, fpvi_end, fpv_conffile, tag ) == -1 ) start_quit_seq();
        fpvi = fpvi_begin;
        while ( fpvi != fpvi_end ) {
    	    strcpy ( buf , *fpvi );
    	    fpvi ++;
    	    strcpy(option, awk(buf, 1));
    	    if (!strcmp(option, "iface")) {
		strcpy( buf2, "" );
	        strcpy( dev , awk(buf, 2) );
		if ( !test->validSysDev(dev)) {
		    error->dump( 16, buf );
		    start_quit_seq();
		}
		sprintf( buf2, "tc qdisc del root dev %s 2> /dev/null > /dev/null", dev );
	        sys->tc( buf2 );
		sprintf( buf2, "tc qdisc add dev %s root handle 1: htb default 1", dev );
		sys->tc( buf2 );
		sprintf( buf2, "tc class add dev %s parent 1: classid 1:1 htb rate 100mbit quantum 8738", dev );
		sys->tc( buf2 );
	    } 
	}
	n++;
    }

    n = 1;

    signal( SIGTERM, signalterm );
    signal( SIGINT, signalterm );
    signal( SIGUSR1, SIG_IGN );

    while ( strlen( awk( sessions, n ))) {
	strncpy( section, "", MAX_NAME );
	strncpy( section, awk(sessions, n), MAX_NAME );
	section_id = 10 + n;
	child_instance.push_back( new Instance( statsunitstr, statsunit ));
	if ( i_am_parent ) {
	    if ( child_instance.back()->getInitStatus() == -1 ) {
		creat( pidfile, 0660 );
  		usleep( 200000 );
  		unlink( pidfile );
    		strcpy( section, "global" );	
		signalterm( SIGTERM );	  
	    }
	} 
	else {
	    parent_instance = child_instance.back(); 
	    if ( strlen(argv[1]) >= 5 ) {
		for ( int i=strlen(argv[1]); i>=0; i-- ) argv[1][i]=0;
		sprintf(argv[1],"child");
	    }
	    else {
		for ( int i=strlen(argv[1]); i>=0; i-- ) argv[1][i]=0;
	    }
	    sprintf(ipt_mychain, "ns_%s", section);
	    break;
	}
	n++;
    }
  
    if  ( i_am_parent ) {
    	chdir("/");

        for ( int i=strlen(argv[1]); i>=0; i-- ) argv[1][i]=0;
	sprintf(buf, "%d", getpid());
	fd = creat(pidfile, 0660);
	write(fd, buf, strlen(buf));
	close(fd);
        strcpy(section, "global");
	info->log((char *)"[start]");
	if ( dump_stats ) {
	    unlink(statspath);
	    fd = creat(statspath, O_WRONLY);
	    sprintf(buf, "chown %s:%s %s && chmod %s %s &", statsfileowner, statsfilegroup, statspath, statsfilemode, statspath);
	    system(buf);
	    close(fd);
	}

	for ( int n=fpv_conffile.size()-1; n>=0; n-- ) { delete [] fpv_conffile.back(); fpv_conffile.pop_back(); }
        for ( int n=fpv_classfile.size()-1; n>=0; n-- ) { delete [] fpv_classfile.back(); fpv_classfile.pop_back(); }

	error->setLogOnTerminal ( false );
	warning->setLogOnTerminal ( false );
	info->setLogOnTerminal ( false );
	
	while (true) {
	    for ( unsigned int n=0; n<child_instance.size(); n++ ) {
		if ( child_instance[n]->getResult() == -1 ) signalterm(SIGTERM);
		child_instance[n]->sendContinue();
		if ( dump_stats ) {
		    fd = open( statspath, O_RDWR | O_TRUNC );
		    for ( unsigned int n=0; n<child_instance.size(); n++ ) child_instance[n]->dumpStats(fd);
		    strcpy( buf, "                                Powered by NiceShaper\n" );
		    write( fd, buf, strlen(buf));
		    strcpy( buf, "                       http://niceshaper.jedwabny.net\n" );
		    write( fd, buf, strlen(buf));
		    close(fd);
		}
	    }
	}
    }
    else {
	unsigned long int cycle;
	chdir("/tmp");
    
	info->setDoNotPutNewLineChar();
        info->dump( 3, "" );
	ns = new NiceShaper();
	if ( ns->init( fpv_conffile, fpv_classfile, protected_marks ) == -1 ) {
	    error->dump( 30, "" );
	    parent_instance->sendInitStatusError();
	    usleep(1000000);
	}
	else {
	    parent_instance->sendInitStatusOk();
        }

	for ( int n=fpv_conffile.size()-1; n>=0; n-- ) { delete [] fpv_conffile.back(); fpv_conffile.pop_back(); }
        for ( int n=fpv_classfile.size()-1; n>=0; n-- ) { delete [] fpv_classfile.back(); fpv_classfile.pop_back(); }

	error->setLogOnTerminal ( false );
	warning->setLogOnTerminal ( false );
	info->setLogOnTerminal ( false );

	gettimeofday( &past, NULL );

        while (true) {
	    usleep (reload_quantum);
	    gettimeofday( &now, NULL );
	    cycle=(now.tv_sec-past.tv_sec)*10+(now.tv_usec-past.tv_usec)/100000;
	    if ( cycle >= ns->reload())
	    {
    		gettimeofday(&past, NULL);
		if ( ns->judge() == -1 )
		{
		    parent_instance->sendFatalError();
		    while (true) // waiting for SIGTERM, from parent instance
		    {
			usleep(1000000);
		    }
		}
		parent_instance->sendResult(ns);
		parent_instance->waitForContinue();
	    }
	    else {
    		parent_instance->sendContinue();
    		parent_instance->waitForContinue();
	    }
	}
    }
    return 0;
}

void run(int argc, char *argv[])
{
    int dpid, fd;
    char buf[LONG_BUF];
    struct timeval now;
    bool force_overwrite_pidfile = false;  
    char usersfile[MAX_PATH]="/etc/niceshaper0.6/users.ns";
    char outfile[MAX_PATH]="";
    char download_section[MAX_PATH]="";
    char upload_section[MAX_PATH]="";
    char iface_inet[MAX_IFACE]="";
    bool resolve_hostname=false;
    vector < char * > fpv;  
    vector < char * >::iterator fpvi, fpvi_end;
    FILE *fp;
    class Conv *conv;

    if (argc==1) { help(); exit(0); }

    for ( int n=2; n<argc; n++ ) {
	if ( string(argv[n]) == "--confdir" ) {
	    sprintf( conffile, "%s/config.ns", argv[n+1] );
	    sprintf( classfile, "%s/class.ns", argv[n+1] );
	    sprintf( usersfile, "%s/users.ns", argv[n+1] );
	}
    }
    
    for ( int n=2; n<argc; n++ ) {
	if ( string(argv[n]) == "--conffile" ) { 
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    strcpy( conffile, argv[++n] ); 
	    if ( !test->fileIsReadable(conffile)) error->dump( 26, conffile );
	}
	else if ( string(argv[n]) == "--classfile" ) { 
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    strcpy( classfile, argv[++n] ); n++; 
	    if ( !test->fileIsReadable(classfile)) error->dump( 26, classfile );
	}
	else if ( string(argv[n]) == "--usersfile" ) { 
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    strcpy( usersfile, argv[++n] ); 
	    if ( !strlen( usersfile )) { error->dump( 28, argv[n] ); exit(0); };
	}
	else if ( string(argv[n]) == "--outfile" ) { 
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    strcpy( outfile, argv[++n] );  
	}
	else if ( string(argv[n]) == "--download-section" ) { 
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    strcpy( download_section, argv[++n] ); 
	}	
	else if ( string(argv[n]) == "--upload-section" ) { 
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    strcpy( upload_section, argv[++n] );  
	}
	else if ( string(argv[n]) == "--iface-inet" ) {
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    strcpy( iface_inet, argv[++n] );	
    	    if ( !test->validSysDev(iface_inet)) { error->dump( 28, argv[n-1] );	exit(0); }
	}
	else if ( string(argv[n]) == "--resolve-hostname" ) { 
	    if ( (n+1)>=argc ) { error->dump( 28, argv[n] ); exit(0); }
	    if ( !strcmp( argv[++n], "true" )) resolve_hostname=true; 
	    else resolve_hostname=false;
	}	
	else if ( string(argv[n]) == "--no-daemon" ) { 
	    daemon_mode=false; 
	}
	else if ( string(argv[n]) == "--force-pidfile" ) { 
	    force_overwrite_pidfile=true; 
	}
	else if ( string(argv[n]) == "--confdir" ) { 	
	    n++;
	}
	else error->dump( 28, argv[n] );
	
	if ( error->fatalError()) exit(-1);
    }

    if ( string(argv[1]) == "convert" ) {
	if ( strlen(outfile) && !access( outfile, 0 )) {
	    gettimeofday( &now, NULL );
	    sprintf( buf, "%s.%lu", outfile, now.tv_sec );
	    if ( lang==pl ) printf( "\n\tplik '%s' juz istnieje, jego nazwa zostanie zmieniona na '%s'\n\n", outfile, buf );
	    else printf( "\n\tfile '%s' already exist, renamed to '%s'\n\n", outfile, buf );
	    rename( outfile, buf );
	}

	if ( test->fileIsReadable(usersfile)) fp=fopen( usersfile, "r" );
	else { error->dump( 26, usersfile ); exit(-1); }
	
	if ( !strlen( iface_inet )) {
            strcpy(iface_inet, "eth0");
            warning->dump( 5, "--iface-inet eth0" );
	}                                               
	if ( !strlen( download_section )) {
	    warning->dump( 6, "--download-section" );
	}
	if ( !strlen( upload_section )) {
	    warning->dump( 6, "--upload-section" );                 
	}           
	
	conv = new Conv;	
	conv->convertUsersFile ( fpv, fp , iface_inet, download_section, upload_section, NULL, resolve_hostname, true );
	fclose(fp);
	delete conv;

        fpvi=fpv.begin();             
        fpvi_end=fpv.end();
        
	if ( strlen(outfile)) {
	    fd = creat( outfile, 0660 );
	    if ( fd == -1 ) { error->dump( 27, outfile ); exit(-1); }
    	    while ( fpvi != fpvi_end ) {	    
		write( fd, *fpvi, strlen(*fpvi) ); 
		write( fd, "\n", 1 ); 
		fpvi++;
	    }	    
	    close(fd);
	}
	else {
	    while ( fpvi != fpvi_end ) {	    
		printf( "%s\n", *fpvi );
		fpvi++;
	    }
	}
	exit(0);		
    } 
    else if (( string(argv[1]) == "stop") || ( string(argv[1]) == "restart" )) {
	if ( access( pidfile, 0 )) {
	    if ( lang==pl ) printf ( "\n\tNiceShaper nie jest uruchomiony...( nie znaleziono pliku /var/run/niceshaper.pid file)\n\n ");
	    else printf ( "\n\tNiceShaper is not running...( there is not /var/run/niceshaper.pid file)\n\n" );
	    exit(0);
	}
	fd = open( pidfile, 0660 );
        dpid = read( fd, buf, 8 );
        buf[dpid] = '\0';
        dpid = atoi(buf);
        close(fd);
        kill( dpid, SIGTERM );
        while ( !access( pidfile, 0 )) 
	{
	    printf(".");
	    fflush(stdout);
	    usleep(100000);
	}
	info->onTerminal((char *)"[stop]");
	if ( !strcmp( argv[1], "stop" )) exit(0);
    } 
    else if ( string(argv[1]) == "stats" ) {
	char buf[10];
	if ( access( pidfile, 0 )) {
	    if ( lang==pl ) printf ( "\n\tNiceShaper nie jest uruchomiony...( nie znaleziono pliku /var/run/niceshaper.pid file)\n\n" );
	    else printf ( "\n\tNiceShaper is not running...( there is not /var/run/niceshaper.pid file)\n\n" );	  
	    exit(0);
	}
	fd = open( pidfile, 0660 );
        dpid = read( fd, buf, 8 );
        buf[dpid] = '\0';
        dpid = atoi(buf);
        close(fd);
        while ( access( statsfile, 0 )) { 
    	    kill( dpid, SIGUSR1 ); 
    	    usleep(100000);
        }
        fd = open( statsfile, 0660 );
        while ( read( fd, buf, 1 )) printf("%c", buf[0]);
        close(fd);
	unlink(statsfile);
        exit(0);
    }
    if (( string(argv[1]) != "start" ) && ( string(argv[1]) != "restart" )) {
	help();
        exit(0);
    }
    if ( !access( pidfile, 0 )) {
        if ( force_overwrite_pidfile ) {
	    unlink(pidfile); 
	}
	else {	
	    if (lang==pl) {
	        printf ( "\n\tNiceShaper jest juz uruchomiony...( znaleziono plik /var/run/niceshaper.pid file) \n" );
	        printf ( "\tmozesz uzyc opcji --force-pidfile jesli wiesz co robisz!!\n\n" );
	    }    
	    else {
		printf ( "\n\tNiceShaper alredy running...( there is /var/run/niceshaper.pid file) \n" );
	        printf ( "\tYou can use --force-pidfile but be carefully!!!\n\n" );	
    	    }		
	    exit(0);
	}
    }
    if ( access( classfile, R_OK )) {
	if ( lang==pl ) printf( "\n\todczytanie pliku %s nie powiodlo sie\nwychodze...\n\n", classfile );
	else printf( "\n\tcan't read %s file\nquit...\n\n", classfile );
	exit(0);
    }
    if ( access( conffile, R_OK )) {
	if ( lang==pl ) printf( "\n\todczytanie pliku %s nie powiodlo sie\nwychodze...\n\n", conffile );	
	else printf( "\n\tcan't read %s file\nquit...\n\n", conffile );	
	exit(0);
    }
    printf( "\n\tNiceshaper version:%s\n\thttp://niceshaper.jedwabny.net\n\n", version );
}

void start_quit_seq() 
{
    error->dump( 30, "" );
    creat(pidfile, 0660);
    usleep(200000);
    unlink(pidfile);
    exit(-1);
}
    
void signalterm( int sig )
{
    signal( SIGTERM, signalterm );

    if ( i_am_parent ) {
	for ( int n=child_instance.size()-1; n>=0; n-- ) {
	    child_instance[n]->killChild();
	    delete child_instance[n];
	}
	child_instance.clear();
	usleep(100000);
	unlink(pidfile);
	info->log((char *)"[stop]");
    } 
    else {
        delete ns;
        delete parent_instance;
	info->dump ( 2, "" );
    }
    exit(0);
  
}

void signalstats( int sig )
{
    int fd;
    
    signal( SIGUSR1, signalstats );

    fd=creat( statstmp, 0660 );
    for ( unsigned int n=0; n<child_instance.size(); n++ ) child_instance[n]->dumpStats(fd);
    close(fd);
    rename (statstmp,statsfile);
}

void help()
{
    printf( "\n\tNiceshaper version:%s\n", version);
    printf( "\thttp://niceshaper.jedwabny.net\n\n");
    printf( "\tusage: niceshaper {start|stop|restart|convert|stats} [OPTIONS]\n\n");
    
    printf( "\tto overwrite default config files locations:\n");
    printf( "\t--conffile PATH-TO-FILE\n");
    printf( "\t--classfile PATH-TO-FILE\n\n");
    printf( "\tto overwrite config dir:\n");
    printf( "\t--confdir PATH-TO-DIR\n\n");
    
    printf( "\tconvert usersfile:\n");
    printf( "\t--usersfile PATH-TO-FILE - usersfile is simple file format from NiceShaper 0.5\n");        
    printf( "\t--outfile PATH-TO-FILE - if not defined classes will be dumped on your terminal\n");    
    printf( "\t--resolve-hostname {true|false}\n\n");     
       
    printf( "\trequired for convert:\n");
    printf( "\t--download-section SECTION - for download configured section\n");
    printf( "\t--upload-section SECTION - for upload configured section\n");
    printf( "\t--iface-inet IFACE - internet interface\n\n");    
    
    printf( "\tother:\n");
    printf( "\t--no-daemon to no act as background proces\n");
    printf( "\t--force-pidfile force run even if there is pidfile\n\n");	
}

