/*                                                                                                                                                                                 
 *      NiceShaper - Dynamic Traffic Management                                                                                                                                    
 *                                                                                                                                                                                 
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>                                                                                                              
 *                                                                                                                                                                                 
 *      This file is subject to the terms and conditions of the GNU General Public                                                                                                 
 *      License.  See the file COPYING in the main directory of this archive for                                                                                                   
 *      more details.                                                                                                                                                              
 */   

#include "net.h"

Net::Net()
{
    IptApplied = false;
    strncpy( IptMyParent, "", SHORT_BUF+1 );
    strncpy( Dev, "", MAX_IFACE+1 );
    strncpy( Match, "", LONG_BUF+1 );
}

Net::~Net()
{
  char buf[LONG_BUF];

  if (!Alias) {
    sprintf(buf, "tc qdisc del root dev %s 2> /dev/null > /dev/null", Dev);
    sys->tc(buf);
  }
  if (IptApplied) {
    sprintf(buf, "iptables -t mangle -D %s %s -j %s", IptMyParent, IptMatch, ipt_mychain);
    sys->ipt(buf);
  }
}
int Net::init(char buf[], bool alias, std::vector < char * > &section_ifaces)
{
    char filter_mode[16];
    Alias = alias;
  
    strncpy ( Match , "", LONG_BUF+1 );
    strncpy ( Dev, awk( buf, 2 ), MAX_IFACE);
    strcpy ( filter_mode, awk( buf, 3 ));
    
    if ( !strcmp( filter_mode, "match" )) {
	strcpy( Match, &(strstr(buf, "match"))[5] );
	if ( trim( Match ) == -1 ) {
	    error->dump( 18, Match );
	    return -1;
        }  
    }
    else if ( !strcmp( filter_mode, "src" )) {
	strcpy( Match, "srcip ");
	strcat( Match, awk( buf, 5 ));
    }
    else if ( !strcmp(filter_mode, "dst" )) {
	strcpy( Match, "dstip " );
	strcat( Match, awk( buf, 5 ));
    }
    else {
	return -1;
    }

    if ( gen_ipt_filter( IptMatch, Match ) == -1 )
    {
	error->dump( 18, Match );
	return -1;
    }
    
    if ( !alias ) {
	section_ifaces.push_back(new char[MAX_IFACE+1]);
	strncpy( section_ifaces.back(), "", MAX_IFACE+1 );
	strncpy( section_ifaces.back(), Dev, MAX_IFACE );
    }

    return 0;
}

int Net::envInit( char ipt_my_parent_mode[], char ipt_my_parent[], unsigned int section_htb_ceil )
{
    char buf[LONG_BUF];
    char tc_rateceil_buf[SHORT_BUF];
 
    strncpy( IptMyParent, ipt_my_parent, SHORT_BUF );
 
    sprintf(buf, "iptables -t mangle %s %s %s -j %s", ipt_my_parent_mode, IptMyParent, IptMatch, ipt_mychain);
    sys->ipt(buf);
    IptApplied = true;

    if ( strstr( Dev, "imq" )) { sprintf(buf, "ip link set %s up", Dev); sys->tc(buf); }
    if ( Alias ) return 0;
	
    // sprintf(buf, "tc class add dev %s parent 1:1 classid 1:%d htb rate %d quantum %d", dev, section_id, section_htb_ceil, compute_quantum(section_htb_ceil));
    // system_tc(buf);
    gen_fixed_tc_rateceil_buf ( tc_rateceil_buf,  section_htb_ceil, section_htb_ceil); 
    sprintf(buf, "tc class add dev %s parent 1:1 classid 1:%d htb %s quantum %d", Dev, section_id, tc_rateceil_buf, compute_quantum(section_htb_ceil));
    sys->tc(buf);

    return 0;
}
