/*
 *      NiceShaper - Dynamic Traffic Management
 *
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>
 *
 *      This file is subject to the terms and conditions of the GNU General Public
 *      License.  See the file COPYING in the main directory of this archive for
 *      more details.
 */

#include "niceshaper.h"

NiceShaper::NiceShaper()
{
    Reload = 50;
    CrossBar = 1;
    SectionHtbCeil = 0;
    SectionShape = 0;
    StatsClasses = c_working;
    StatsSum = s_bottom;
    strncpy( IptMyParent, "" , SHORT_BUF+1);
    strncpy( IptMyParentMode, "" , SHORT_BUF+1);
}

NiceShaper::~NiceShaper()
{
    char buf[LONG_BUF];

    for ( int n=net.size()-1; n>=0; n-- ) { delete net.back(); net.pop_back(); }
    for ( int n=tcclass.size()-1; n>=0; n-- ) { delete tcclass.back(); tcclass.pop_back(); }

    sprintf(buf, "iptables -t mangle -F %s 2> /dev/null > /dev/null", ipt_mychain);
    sys->ipt(buf);
    sprintf(buf, "iptables -t mangle -X %s 2> /dev/null > /dev/null", ipt_mychain);
    sys->ipt(buf);
}

int NiceShaper::init(vector < char * > &fpv_conffile, vector < char * > &fpv_classfile, vector < unsigned int > &protected_marks )
{
    unsigned int default_low = 0;
    unsigned int default_ceil = 0;
    unsigned int default_htb_prio = 5;
    unsigned int default_hold = 30;
    unsigned int default_prio = 0;
    unsigned int default_strict = 70;
    unsigned int classid = 1000;
    unsigned int filterid = 0;
    unsigned int sfq_perturb = 0;    
    unsigned int esfq_perturb = 0;
    unsigned int section_classes = 0;    
    char classtype[SHORT_BUF] = "standard-class";
    char scheduler[SHORT_BUF] = "sfq";
    char esfq_hash[SHORT_BUF] = "";
    char buf[LONG_BUF], var_buf[LONG_BUF];
    char option[SHORT_BUF+1] = "", param[SHORT_BUF+1] = "", value[SHORT_BUF+1] = "";
    char used_for_fw_marks[LONG_BUF];
    bool alias;
    bool imq_auto_redirect = true;
    char section_set[SHORT_BUF];
    char section_end[SHORT_BUF];
    vector < char * >::iterator fpvi, fpvi_begin, fpvi_end, fpvi_tmp; 
    vector < char * > fpv_myclasses;
    char *buf_ptr;  
    bool mydatablock ;
    unsigned int configured_class_default_mark;
    unsigned int auto_class_default_mark;
    bool is_protected;  
    bool default_param_should_be_inserted;
    ipt_targettype ipt_target = t_accept;
    TcClass *iterclass, *tpltcclass;    

    sprintf( section_set, "<%s>", section );
    sprintf( section_end, "</%s>", section );
    
    
    /* reading global config */
    if ( fpv_section_i( fpvi_begin, fpvi_end, fpv_conffile, "global" ) == -1 ) 
	return -1;

    /*fpvi = fpvi_begin;
    while ( fpvi != fpvi_end ) {
	printf("cf:%s\n",*fpvi);
        fpvi++;
    }*/

    fpvi=fpvi_begin;
    while ( fpvi != fpvi_end ) {
	strncpy ( buf , *fpvi, LONG_BUF );
	fpvi++;
	strncpy ( option, awk( buf, 1 ), SHORT_BUF );
        strncpy ( param, awk( buf, 2 ), SHORT_BUF );                                                                                                                              
        strncpy ( value, awk( buf, 3 ), SHORT_BUF );
	if ( !strcmp( option, "mark-on-ifaces" )) 
	{
            MarkIfaces.push_back(new char[MAX_IFACE+1]);
            strncpy( MarkIfaces.back(), "", MAX_IFACE+1 );
            strncpy( MarkIfaces.back(), param, MAX_IFACE );
        }
	else if ( !strcmp( option, "stats" ))
	{
    	    if ( !strcmp( param, "classes" ))
    	    {
    		if ( !strcmp( value, "all" )) StatsClasses = c_all;
    		else if ( !strcmp( value, "active" )) StatsClasses = c_active;
    		else if ( !strcmp( value, "working" )) StatsClasses = c_working;
    		else if ( !strcmp( value, "false" )) StatsClasses = c_false;
		else { error->dump( 11, buf ); }
	    }
    	    else if ( !strcmp( param, "sum" ))
    	    {
    		if ( !strcmp( value, "top" )) StatsSum = s_top;
    		else if ( !strcmp( value, "bottom" )) StatsSum = s_bottom;
    		else if ( !strcmp( value, "false" )) StatsSum = s_false;
    		else { error->dump( 11, buf ); }
    	    }
	}
    }

    /* reading section config */
    if ( fpv_section_i( fpvi_begin, fpvi_end, fpv_conffile, section ) == -1 ) {
	return -1;
    }
    
    /*fpvi = fpvi_begin;
    while ( fpvi != fpvi_end ) {
	printf("cf:%s\n",*fpvi);
	fpvi++;
    }*/

    fpvi = fpvi_begin;  
    while ( fpvi != fpvi_end ) 
    {
	strcpy ( buf , *fpvi );

	strncpy ( option, awk( buf, 1 ), SHORT_BUF);
	strncpy ( param, awk( buf, 2 ), SHORT_BUF);
	strncpy ( value, awk( buf, 3 ), SHORT_BUF);
 	if ( !strcmp( option, "low" )) {
	    default_low = unitconvert( param, bit);
	    if ( default_low > 1073741824 ) error->dump( 32, buf ); 
	}     
	else if ( !strcmp( option, "ceil" )) {
	    default_ceil = unitconvert( param, bit);
    	    if ( default_ceil > 1073741824 ) error->dump( 32, buf ); 
	}     
	else if ( !strcmp( option, "rate" )) {
	    default_low = default_ceil = unitconvert(  param, bit); 
	    if ( default_low > 1073741824 ) error->dump( 32, buf ); 
	}    
	else if ( !strcmp( option, "hold" )) 
	    default_hold = atoi( param ); 
	else if ( !strcmp( option, "prio" )) 
	    default_prio = atoi( param ); 
	else if ( !strcmp( option, "strict" )) 
	    default_strict = atoi( param );
	else if ( !strcmp( option, "imq" )) {
	    if ( !strcmp( param, "autoredirect" )) {
		if ( !strcmp( value, "true" )) imq_auto_redirect = true;
		else if ( !strcmp( value, "false" )) imq_auto_redirect = false;
		else { error->dump( 11, buf ); }	
	    }
	    else { error->dump( 11, buf ); }	    
	}    
    	else if ( !strcmp( option, "do-not-shape" )) 
	    strcpy(classtype, "do-not-shape");
    	else if ( !strcmp( option, "virtual" )) 
	    strcpy(classtype, "virtual");
    	else if ( !strcmp( option, "wrapper" )) 
	    strcpy(classtype, "wrapper");
	else if ( !strcmp( option, "section" )) {
	    if ( !strcmp( param, "speed" )) {
		SectionHtbCeil = unitconvert( value, bit );
		if ( SectionHtbCeil > 1073741824 ) error->dump( 32, buf ); 
	    }
	    else if ( !strcmp( param, "shape" )) {
		SectionShape = unitconvert( value, bit );
		if ( SectionShape > 1073741824 ) error->dump( 32, buf ); 
	    }
	    else { error->dump( 11, buf ); }
	} 
	else if ( !strcmp( option, "htb" )) {
	        if ( !strcmp( param, "prio" )) default_htb_prio = atoi( value ); 
		else if ( !strcmp( param, "scheduler" )) strcpy( scheduler, value ); 
	        else error->dump( 11, buf ); 
	}    
	else if ( !strcmp( option, "sfq" )) {
	    if ( !strcmp( param, "perturb" )) sfq_perturb = atoi( value );
	    else error->dump( 11, buf ); 
	}
	else if ( !strcmp( option, "esfq" )) {                                                                                                                   
	    if ( !strcmp( param, "hash" )) {                                                                                                                
		if ( !strcmp( value, "classic" )) strcpy( esfq_hash, "classic" );                                                                                
		else if ( !strcmp( value, "dst" )) strcpy( esfq_hash, "dst" );                                                                                   
		else if ( !strcmp( value, "src" )) strcpy( esfq_hash, "src" );                                                                                   
		else error->dump( 11, buf );                                                                                                        
	    }   
	    else if ( !strcmp( param, "perturb" ))                                                                                                                 
		esfq_perturb = atoi( value );                                                                                                                    
	    else 
		error->dump( 11, buf );                                                                                                        	    
	}                                                                                                                                                 
	else if ( !strcmp( option, "reload" )) 
	{  
	    if ( !strcmp( param, "false" )) {
    		Reload = 0;
	    }
	    else {
    		Reload = (unsigned int)(strtof(param, NULL)*10);
    	        if (( Reload == 0 ) || ( Reload > 6000 )) {
		    error->dump( 13, buf );
    		    return -1;
		}
	    }
	}
	else if ( !strcmp( option, "iptables" )) 
	{
	    if (( !strcmp( param, "hook" )) || (!strcmp( param, "append-hook" ))) { 
		strncpy( IptMyParentMode , " -A ", SHORT_BUF ); 
		strncpy( IptMyParent, value, SHORT_BUF ); 
	    }	 	 
	    else if ( !strcmp( param, "insert-hook" )) { 
		strncpy( IptMyParentMode , " -I ", SHORT_BUF ); 
		strncpy( IptMyParent, value, SHORT_BUF ); 
	    }
	    else if ( !strcmp( param, "target" )) { 
	        if ( !strcmp( value, "accept" )) ipt_target = t_accept;
	        else if ( !strcmp( value, "return" )) ipt_target = t_return;
	        else if ( !strcmp( value, "drop" )) ipt_target = t_drop;
	        else { error->dump( 11, buf ); }
	    }	        
	    else { error->dump( 11, buf ); }
	}
	else if ( !strcmp( option, "mode" ))
	{
	    warning->dump( 7, buf );
	    if ( !strcmp( param, "download" )) strncpy( IptMyParent, "POSTROUTING", SHORT_BUF );
	    else if ( !strcmp( param, "upload" )) strncpy( IptMyParent, "PREROUTING", SHORT_BUF );
	    else { error->dump( 14, buf ); return -1; }
	    strncpy ( IptMyParentMode , " -A ", SHORT_BUF );
	}
	else if ( !strcmp( option, "debug" )) 
	{
	    if ( !strcmp( param, "iptables" )) sys->setDebugIpt(true);
	    else if ( !strcmp( param, "iproute" )) sys->setDebugTc(true);
	} 
	else if ( !strcmp( option, "iface" )) 
	{
	    char dev_[MAX_IFACE], dev__[MAX_IFACE];

	    strcpy( dev_, awk( buf, 2 ));
	    fpvi_tmp = fpvi_begin;
	    alias = false;
	    while ( fpvi_tmp != fpvi ) 
	    { 	  
		strcpy ( var_buf , *fpvi_tmp );		
		if ( !strcmp( awk( var_buf, 1 ), "iface" ) && ( fpvi_tmp != fpvi )) 
		{
		    strcpy( dev__, awk( var_buf, 2 ));
	  	    if ( !strcmp( dev_, dev__ )) alias = true;
		}
		fpvi_tmp++;
	    }

	    net.push_back( new Net());
	    if ( net.back()->init( buf, alias, SectionIfaces ) == -1 ) return -1;
	    alias = false;
	}

	if ( error->fatalError()) return -1;
	
	fpvi++;
    }
    
    if ( !SectionHtbCeil ) error->dump(19, ""); 
    if ( !SectionShape ) error->dump(20, ""); 
    if ( !strlen ( IptMyParent ) ) error->dump(21, ""); 
    if ( !net.size() ) error->dump(22, "");
    if ( !default_ceil ) default_ceil = SectionShape; 
    if ( SectionShape > SectionHtbCeil ) error->dump(40, "");
    if ( default_ceil > SectionShape ) default_ceil = SectionShape;
    if ( default_low > default_ceil ) default_low = default_ceil;    

    if ( error->fatalError()) return -1;

/*    fpvi = fpv_classfile.begin();
    while ( fpvi != fpv_classfile.end()) { 
	printf("cl:%s\n",*fpvi);
	fpvi++;
    }*/
    
    mydatablock = false;
    fpvi=fpv_classfile.begin();
    while ( fpvi != fpv_classfile.end()) 
    {
	strcpy( option, awk( *fpvi, 1 ));

	if ( !strcmp( option, "class" )) 
	{
	    classid++;	
	    configured_class_default_mark=0;
	    auto_class_default_mark=0;
	    strcpy( used_for_fw_marks, " " );
	    
	    if ( !strcmp( awk( *fpvi, 2 ), section )) 
	    {
		mydatablock = true;
		
		buf_ptr = new char[LONG_BUF];
		strcpy( buf_ptr, *fpvi );
		fpv_myclasses.push_back( buf_ptr );
		fpvi_tmp = fpvi;
		do {
		    if ( !strcmp( awk( *fpvi_tmp, 1 ), "set-mark" )) configured_class_default_mark=read_fw_mark(awk( *fpvi_tmp, 2 )); 
		    fpvi_tmp++;		    
		} while (( fpvi_tmp != fpv_classfile.end()) && ( strcmp( awk( *fpvi_tmp, 1 ), "class" )));
	    } 
	    else 
	    {
	        mydatablock = false;
	    }
	}	
	else if ( !strcmp( option, "match" )) 
	{
	    filterid++;	  	
    	    do {
	        is_protected = false;
	        for ( unsigned int n = protected_marks.size() ; n ; n-- ) 
	        {
	            if ( filterid == protected_marks[n-1] ) { is_protected = true; filterid++; }
	        }
	    } while ( is_protected );
	
	    if ( !mydatablock ) { fpvi++; continue; }

	    if ( !auto_class_default_mark ) auto_class_default_mark=filterid;

	    buf_ptr = new char[LONG_BUF];
	    sprintf ( buf,"%d",filterid );
	    strcpy ( buf_ptr , *fpvi );
	    strcat ( buf_ptr , " filterid " );
	    strcat ( buf_ptr , buf );

	    if ( configured_class_default_mark )
	    {
		if ( !strstr( *fpvi, " set-mark " ))
		{
		    sprintf( buf, " %d ", configured_class_default_mark );
		    strcat( buf_ptr, " set-mark " );
		    strcat( buf_ptr, buf );
		    if ( !strstr( used_for_fw_marks, buf )) {
			strcat( buf_ptr, " use-for-fw true " );
			strcat( used_for_fw_marks, buf );
		    }	
		}
		else 
		{
	    	    strcpy( buf, awk(strstr( *fpvi, " set-mark " ) ,2 ));
		    if ( !strstr( used_for_fw_marks, buf )) {
			strcat( buf_ptr, " use-for-fw true " );
			strcat( used_for_fw_marks, " " );
			strcat( used_for_fw_marks, buf );
			strcat( used_for_fw_marks, " " );
		    }	
		}
	    }
	    else if ( strstr( *fpvi, " mark " ))
	    {
		sprintf( buf, " %s ", awk( strstr( *fpvi, " mark " ), 2 ));
		strcat( buf_ptr, " set-mark " );
		strcat( buf_ptr, buf );
		if ( !strstr( used_for_fw_marks, buf )) {
		    strcat( buf_ptr, " use-for-fw true " );
		    strcat( used_for_fw_marks, buf );
		}	
	    }
	    else {
		if ( !strstr( *fpvi, " set-mark " ))
		{
		    sprintf ( buf," %d ", auto_class_default_mark );
		    strcat ( buf_ptr, " set-mark " );
		    strcat ( buf_ptr, buf );
		    if ( !strstr( used_for_fw_marks, buf )) {
			strcat( buf_ptr, " use-for-fw true " );
			strcat( used_for_fw_marks, buf );
		    }	
		}
		else 
		{
	    	    strcpy( buf, awk( strstr( *fpvi, " set-mark " ), 2));
		    if ( !strstr( used_for_fw_marks, buf )) {
			strcat( buf_ptr, " use-for-fw true " );
			strcat( used_for_fw_marks, " " );
			strcat( used_for_fw_marks, buf );
			strcat( used_for_fw_marks, " " );
		    }	
		}

	    }
	    
	    fpv_myclasses.push_back(buf_ptr);
	    
	    fpvi_tmp=fpvi;
	    fpvi_tmp++;
	    
	    if ( fpvi_tmp == fpv_classfile.end()) default_param_should_be_inserted = true;
	    else if ( strcmp( awk(*fpvi_tmp, 1), "match")) default_param_should_be_inserted = true;
	    else default_param_should_be_inserted = false;
	     
	    if ( !default_param_should_be_inserted ) { fpvi++; continue; }
	    
	    buf_ptr = new char[LONG_BUF];
            sprintf( buf_ptr, "%s", classtype );
            fpv_myclasses.push_back( buf_ptr );
            
            buf_ptr = new char[LONG_BUF];
            sprintf( buf_ptr, "classid %d", classid );
            fpv_myclasses.push_back( buf_ptr );
                                     	    
    	}
	else if ( mydatablock )
	{
	    buf_ptr = new char[LONG_BUF];
	    strcpy ( buf_ptr , *fpvi );
	    fpv_myclasses.push_back( buf_ptr );
	}
	fpvi++;
    }

    if ( !fpv_myclasses.size() ) { error->dump( 23, "" ); return -1; }

    /* Initialize template object with default values */

    tpltcclass = new TcClass( SectionIfaces, MarkIfaces );
    
    sprintf( buf, "low %d", default_low );
    tpltcclass->store( buf );
	    
    sprintf( buf, "ceil %d", default_ceil );
    tpltcclass->store( buf );

    sprintf( buf, "strict %d", default_strict );
    tpltcclass->store( buf );
    	    
    sprintf( buf, "prio %d", default_prio );
    tpltcclass->store( buf );

    sprintf( buf, "htb prio %d", default_htb_prio );
    tpltcclass->store( buf );

    sprintf( buf, "htb scheduler %s", scheduler );
    tpltcclass->store( buf );

    sprintf( buf, "hold %d", default_hold );
    tpltcclass->store( buf );

    if ( sfq_perturb ) {
        sprintf( buf, "sfq perturb %d", sfq_perturb );
	tpltcclass->store( buf );
    }	

    if ( esfq_perturb ) {
        sprintf( buf, "esfq perturb %d", esfq_perturb );
	tpltcclass->store( buf );
    }	

    if ( strlen ( esfq_hash ) ) {
	sprintf( buf, "esfq hash %s", esfq_hash );
	tpltcclass->store( buf );
    }	
	
    if ( imq_auto_redirect ) 
    {
        strcpy ( buf, "imq autoredirect true" );
        tpltcclass->store( buf );
    }
    else  
    {
        strcpy ( buf, "imq autoredirect false" );
        tpltcclass->store( buf );
    }
	
    if ( ipt_target == t_accept )
    {
	strcpy ( buf, "iptables target accept" );
        tpltcclass->store( buf );
    }
    else if ( ipt_target == t_return )
    {
	strcpy ( buf, "iptables target return" );
        tpltcclass->store( buf );
    }
    else if ( ipt_target == t_drop )
    {
	strcpy ( buf, "iptables target drop" );
        tpltcclass->store( buf );
    }
	    
    /* Initialize TcClass objects, using above template and extra parameters */	    
    fpvi=fpv_myclasses.begin();
    do {
	strcpy( option, awk(*fpvi, 1));
        if ( !strcmp( option , "class" )) {
            section_classes++;
            tcclass.push_back(new TcClass(*tpltcclass));
        }
	iterclass=tcclass.back();
        if ( iterclass->store(*fpvi) == -1 ) return -1;
        fpvi++;	
    } while ( fpvi != fpv_myclasses.end() );
  
    delete tpltcclass;  
  
    for ( unsigned int i = 0; i < tcclass.size(); i++ ) {
    	iterclass=tcclass[i];
	if ( iterclass->prepareTcClass( SectionShape, section_classes ) == -1 ) return -1;
	if ( iterclass->testParams() == -1 ) return -1;	
    }
    gettimeofday(&Ltt, NULL);
    sprintf(buf, "iptables -t mangle -N %s", ipt_mychain);
    sys->ipt(buf);
  
    for (unsigned int i = 0; i < tcclass.size(); i++) {
	if (tcclass[i]->envInit() == -1) return -1;
	if ( !(i%10)) 
	{ 
	    info->setDoNotPutTabChar();
	    info->setDoNotPutNewLineChar();
	    info->onTerminal((char *)".");
	}
    }
    info->onTerminal((char *)"");

    for (unsigned int i = 0; i < net.size(); i++) { 
	if (net[i]->envInit( IptMyParentMode, IptMyParent, SectionHtbCeil ) == -1) return -1; 
    }
    
    for ( int n=fpv_myclasses.size()-1; n>=0; n-- ) { delete [] fpv_myclasses.back(); fpv_myclasses.pop_back(); }
    return 0;
}

int NiceShaper::checkIptables ()
{
    vector < char * > iptables_summary;      
    struct timeval now;    
    float cycle_time;
    char buf[LONG_BUF];    
    FILE *fp;  
    TcClass * iterclass;    

    gettimeofday(&now, NULL);
    cycle_time =  (float) (1000000 * (now.tv_sec - Ltt.tv_sec) + now.tv_usec - Ltt.tv_usec) / 1000000;
    gettimeofday(&Ltt, NULL);
    sprintf(buf, "iptables -t mangle -L -Z %s -vnx", ipt_mychain);
    fp = popen(buf, "r");
    for ( unsigned int n=1; n<=2; n++ ) {
        if ( fgets(buf, LONG_BUF, fp) == NULL ) {
	    error->dump( 12, "" ); 
    	    pclose(fp);
	    return -1;
	}
    }
    
    char * bufptr;
    while ( fgets( buf, LONG_BUF, fp )) {
	bufptr = new char[LONG_BUF];
        strcpy ( bufptr , buf );
        iptables_summary.push_back(bufptr);
    }
    pclose(fp);

    delete [] iptables_summary.back();
    iptables_summary.pop_back();
    
    SectionRate = 0; 
    ChangeableSectionRate = 0; 
    MaxPotentialSectionRate = 0; 
    Working = 0; 
    for ( int i = tcclass.size()-1; i >= 0 ; i-- ) {
	iterclass=tcclass[i];
	iterclass->check( iptables_summary, now, cycle_time );
	if ( iterclass->work() && iterclass->sumInSection())
	{
	    WorkingClass.push_back( iterclass );
	    SectionRate+=iterclass->rate();
	    MaxPotentialSectionRate+=iterclass->htbCeil();
	    Working++;
	    if ((iterclass->rate() > iterclass->nsLow()) && (iterclass->rate() <= iterclass->nsCeil()))
	    {
		ChangeableClass.push_back( iterclass );
		ChangeableSectionRate += (iterclass->rate() - iterclass->nsLow());
	    }
	}
    }

    SectionMeasureRate = SectionRate;

    if ( iptables_summary.size()) { error->dump ( 12, "" ); return -1; }

    return 0;
}

int NiceShaper::judge ()
{
    unsigned int old_max_potential_section_rate = 0;
    unsigned int cycle_counter = 0;
    unsigned int sumofw = 0, sumofc = 0;
    unsigned int topbarbits;
    unsigned int alfa = 100;
    bool overload; 
    bool break_conditions;
    bool low_instead_softlimit = false;
    register int i = 0;
    TcClass *iterclass;   

    if ( WorkingClass.size()) WorkingClass.clear();
    if ( ChangeableClass.size()) ChangeableClass.clear();

    if ( checkIptables() == -1 ) return -1;

    if ( WorkingClass.empty())
	return 0; 

    for ( i=WorkingClass.size()-1; i>=0 ; i-- ) {
        iterclass=WorkingClass[i];
	sumofw+=(iterclass->htbCeil()-iterclass->nsLow());
    }

    for ( i=ChangeableClass.size()-1 ; i>=0 ; i-- ) {
        iterclass=ChangeableClass[i];	
	sumofc+=(iterclass->htbCeil()-iterclass->nsLow());
    }

    if ( !sumofw ) return 0;
     
    if ( SectionRate > SectionShape ) {
	topbarbits = 1024*(unsigned int)(((double)SectionShape*CrossBar)/(SectionHtbCeil/working));
	CrossBar -= (double)topbarbits/sumofw;
	if ( CrossBar < 0 ) CrossBar = 0;
	overload = true;
    }
    else {
	topbarbits = (SectionShape-SectionRate)/working; 
	CrossBar += (double)topbarbits/sumofw;
	if ( CrossBar > 1 ) CrossBar = 1;	    
	overload = false;	    
    }

    for ( i=WorkingClass.size()-1; i>=0; i-- ) {
	iterclass=WorkingClass[i];	
	iterclass->setHardLimit( iterclass->nsLow() + (unsigned int)((double)( iterclass->nsCeil()-iterclass->nsLow()) * CrossBar));
	iterclass->computeSoftLimit(); 
	if ( iterclass->htbCeil() > iterclass->hardLimit()) {
    	    sumofw -= ( iterclass->htbCeil() - iterclass->hardLimit());
	    if ( iterclass->rate() > iterclass->nsLow() ) { 
		sumofc -= ( iterclass->htbCeil() - iterclass->hardLimit());
	    }
	    iterclass->setHtbCeil( iterclass->hardLimit());
	    if ( iterclass->rate() > iterclass->htbCeil() ) {
		SectionRate -= (iterclass->rate() - iterclass->htbCeil());
	        iterclass->setRate( iterclass->htbCeil()); 
	    }
	}
    }

    if ( ChangeableClass.empty()) alfa = 100;
    else if ((alfa = 1000/ChangeableClass.size()) > 100 ) alfa = 100;

    if ( ChangeableClass.empty())
    {
	if ( !overload ) {
	    for ( int i=WorkingClass.size()-1; i>=0; i-- ) 
	    {
		iterclass=WorkingClass[i];
		iterclass->incHtbCeil ((unsigned int)(( SectionShape - SectionRate ) * ((double)(iterclass->htbCeil()-iterclass->nsLow())/sumofw)));
		if ( iterclass->htbCeil() == iterclass->nsLow()) iterclass->incHtbCeil( alfa );
    	    	if ( iterclass->htbCeil() > iterclass->nsCeil() ) iterclass->setHtbCeil( iterclass->nsCeil()); 
    	    	
	    }
    	    applyChanges();
	}

        return 0;
    }

    do {
    	if ( cycle_counter ) {
    	    SectionRate = 0;
    	    ChangeableSectionRate = 0;
    	    sumofc=0; sumofw=0;
	    for ( int i=WorkingClass.size()-1; i>=0; i-- ) {
		iterclass=WorkingClass[i]; 
		if ( iterclass->rate() > iterclass->htbCeil()) 
		    iterclass->setRate( iterclass->htbCeil());
		if ((iterclass->rate() > iterclass->nsLow()) && (iterclass->rate() <= iterclass->nsCeil()))
		{
		    ChangeableSectionRate += (iterclass->rate() - iterclass->nsLow());
		    sumofc += (iterclass->htbCeil()-iterclass->nsLow());
		}
		SectionRate += iterclass->rate();
		sumofw += (iterclass->htbCeil()-iterclass->nsLow());

	    }
	    if ( SectionRate < SectionShape )	break; 
	    if ( !sumofc || !sumofw ) break;
	}
    	
        old_max_potential_section_rate=MaxPotentialSectionRate;
	MaxPotentialSectionRate=0; 
	for ( i=WorkingClass.size()-1; i>=0; i-- ) {
	    iterclass=WorkingClass[i];	

	    if ( overload ) {
		if (( iterclass->rate() > iterclass->softLimit()) || ( low_instead_softlimit && ( iterclass->rate() > iterclass->nsLow()))) {
		    iterclass->decHtbCeil((unsigned int)(( SectionRate - SectionShape ) * ((double)( iterclass->htbCeil()-iterclass->nsLow())/sumofc))); 
		    iterclass->decHtbCeil( alfa );
		}
	    }
	    else {
		iterclass->incHtbCeil ((unsigned int)(( SectionShape - SectionRate ) * ((double)(iterclass->htbCeil()-iterclass->nsLow())/sumofw)));
		if ( iterclass->htbCeil() == iterclass->nsLow()) iterclass->incHtbCeil( alfa );
	    }

	    if ( iterclass->htbCeil() > iterclass->nsCeil() ) iterclass->setHtbCeil( iterclass->nsCeil()); 
	    if ( iterclass->htbCeil() < iterclass->nsLow() ) iterclass->setHtbCeil( iterclass->nsLow()); 
	    MaxPotentialSectionRate+=iterclass->htbCeil();
	}
	
	cycle_counter++;

	if ( cycle_counter == 100 ) break_conditions = true;	
	else if (( overload ) && ( MaxPotentialSectionRate == old_max_potential_section_rate ) && ( ChangeableSectionRate > 1000 )) { low_instead_softlimit = true; break_conditions = false; } 
	else if (( overload ) && ( MaxPotentialSectionRate < old_max_potential_section_rate )) { low_instead_softlimit = false; break_conditions = false; }
	else break_conditions = true;	
    } while (!break_conditions);

    applyChanges();

    if ( error->fatalError()) { return -1; }    
    else return 0;
}

int NiceShaper::applyChanges()
{
    for ( int i=WorkingClass.size()-1; i>=0 ; i-- ) {
        WorkingClass[i]->applyChanges();
    }
    
    return 0;
}

void NiceShaper::stats( int pipe_write )
{
    char buf[STAT_MESSAGE_BUF];
    char _buf_;
    unsigned int tcclasssize=tcclass.size();

    sprintf( buf, "%s rate prev-utilize\n", section );
    _buf_ = PROTO_BASE + strlen(buf);
    write( pipe_write, &_buf_, 1 ); 
    write( pipe_write, buf, strlen(buf));

    if ( StatsSum != s_false )
    {
	sprintf( buf, "sum(classes:%d) - %d\n", working, SectionMeasureRate );
	_buf_ = PROTO_BASE + strlen(buf);
    }	
    
    if ( StatsSum == s_top ) 
    {
	write( pipe_write, &_buf_, 1 ); 
	write( pipe_write, buf, strlen(buf));    
    }

    if ( StatsClasses != c_false )
    {
	for ( unsigned int i=0; i<tcclasssize; i++ ) {
	    tcclass[i]->stats( pipe_write, StatsClasses );
        }
    }
    
    if ( StatsSum == s_bottom ) 
    {
	write( pipe_write, &_buf_, 1 ); 
	write( pipe_write, buf, strlen(buf));    
    }
}

