/*                                                                                                                                                                                 
 *      NiceShaper - Dynamic Traffic Management                                                                                                                                    
 *                                                                                                                                                                                 
 *      Copyright 2004 Mariusz Jedwabny <mariusz@jedwabny.net>                                                                                                              
 *                                                                                                                                                                                 
 *      This file is subject to the terms and conditions of the GNU General Public                                                                                                 
 *      License.  See the file COPYING in the main directory of this archive for                                                                                                   
 *      more details.                                                                                                                                                              
 */   

#include "tests.h"

bool Tests::getSysDevices()
{
    struct ifreq *ifr;
    struct ifreq ifr2;
    struct ifconf ifc;
    unsigned int query_devs=0;
    int sd;

    sd = socket (PF_INET, SOCK_STREAM, 0);

    do {
	if (query_devs) delete [] ifr;
	query_devs++;
	ifr = new ifreq[query_devs];
        ifc.ifc_len = query_devs * sizeof (struct ifreq);
	ifc.ifc_req = ifr;
	ioctl (sd, SIOCGIFCONF, &ifc);
    } while (query_devs == ifc.ifc_len / sizeof(struct ifreq));
     
    for (unsigned int i = 0; i < (ifc.ifc_len / sizeof (struct ifreq)); i++) {
	sysNetDevices.push_back(string(ifr[i].ifr_name));
    }
    
    for ( int n=0; n<=15; n++) {
	sprintf( ifr2.ifr_name, "imq%d", n);
	if ( ioctl (sd, SIOCGIFINDEX, &ifr2) < 0 ) {
	    break;
	}
	
	if ( ioctl(sd, SIOCGIFFLAGS, &ifr2) < 0 ) {
	    break;
	}
		    
	if (ifr2.ifr_flags & IFF_UP) {
	    sysNetDevices.push_back(string(ifr2.ifr_name));
        }
    } 

    close(sd);

    return true;
}

bool Tests::validSysDev(char dev[])
{
    vector < string >::iterator fpvi;
    
    fpvi=sysNetDevices.begin();
    while ( fpvi != sysNetDevices.end() ) {
	if (*fpvi==string(dev)) return true;
	fpvi++;
    }
    
    return false;
}

bool Tests::validDev(char dev[])
{
    unsigned int i = 0;
    bool dot = false;
    
    if ( dev[0]==101 && dev[1]==116 && dev[2]==104 ) i=3; // eth
    else if ( dev[0]==97 && dev[1]==116 && dev[2]==104 ) i=3;	// ath    
    else if ( dev[0]==112 && dev[1]==112 && dev[2]==112 && dev[3]==43 && dev[4]==0 ) return true; // ppp+
    else if ( dev[0]==112 && dev[1]==112 && dev[2]==112 ) i=3; // ppp
    else if ( dev[0]==105 && dev[1]==109 && dev[2]==113 ) i=3; // imq
    else if ( dev[0]==98 && dev[1]==114 ) i=2;			// br
    else if ( dev[0]==119 && dev[1]==108 && dev[2]==97 && dev[3]==110 ) i=4; // wlan
    else if ( dev[0]==98 && dev[1]==111 && dev[2]==110 && dev[3]==100 ) i=4; // bond
    else return false;

    if ( dev[i]<48 || dev[i]>57 ) return false;

    i++; 
  
    while ( dev[i]!=0 && dev[i]!=9 && dev[i]!=10 && dev[i]!=32 ) {
	if ( dev[i] == 46 ) {
	    if ( dot == true ) return false;
    	    else dot = true;
	}
	else if (( dev[i]<48 || dev[i]>57 )) {
	    return false; 
        }  
        i++;
	if ( i==MAX_IFACE ) return false;
    }
    
    return true;
}

bool Tests::validIp(char ipaddr[])
{
    struct in_addr inaddr;
    if ( !inet_aton( ipaddr, &inaddr )) return false;
    if ( strcmp( ipaddr, inet_ntoa( inaddr ))) return false;
    return true;
}

bool Tests::validPort(int tcpport)
{
    if (tcpport >= 1 && tcpport <= 65536) return true;
    else return false;
}

bool Tests::validWhiteLine(char buf[])
{
    if (!strlen(buf)) return true;
    else return false;
}

bool Tests::solidIpMask(char ipmask[])
{
    unsigned long addr;

    if ( !validIp( ipmask )) return false;
    addr = htonl( inet_addr( ipmask ));
    for ( int n=1; n<32; n++ )
	if ((( addr>>n)&1)==0 && (( addr>>(n-1))&1)==1 ) return false;
	
    return true;
}

bool Tests::fileIsReadable ( char path[MAX_PATH] ) 
{
    if ( access( path, R_OK ) == -1 ) {
	return false;	
    }

    return true;
}

bool Tests::wordInCharTable ( char buf[], char arg[] ) 
{
    unsigned int n = 0;
    
    while ( strlen(awk( buf, ++n ))) {
	if ( !strcmp( awk( buf, n ), arg )) return true; 
    }
    return false;
}



